#!/etc/ConsolePi/venv/bin/python3
from __future__ import annotations
import os
from typing import Any, Dict
import yaml
import json
import shutil
from pathlib import Path

from consolepi import utils, log  # type: ignore
LOG_FILE = '/var/log/ConsolePi/consolepi.log'

# overridable defaults (via OVERRIDES section of ConsolePi.yaml)
DEFAULT_BAUD = 9600
DEFAULT_DBITS = 8
DEFAULT_PARITY = 'n'
DEFAULT_FLOW = 'n'
DEFAULT_SBITS = 1
DEFAULT_REMOTE_TIMEOUT = 3
DEFAULT_DLI_TIMEOUT = 7
DEFAULT_SO_TIMEOUT = 3  # smart outlets
DEFAULT_CYCLE_TIME = 3
DEFAULT_API_PORT = 5000


class RemoteTimeout:
    def __init__(self, default: int = DEFAULT_REMOTE_TIMEOUT) -> None:
        self.default: int = default

    def add_host(self, host: str, timeout: int):
        setattr(self, host, timeout)

    def __getattr__(self, name: str) -> int:
        name = name.split("(")[0]  # provided hostname(ip-address) from remotes.get_adapters_via_api()
        if name in self.__dict__.keys():
            return getattr(self, name)
        else:
            return self.default


class Config():
    '''Config object contains all statically defined variables and data from config files.'''
    def __init__(self):
        self.static = self.get_config_all('/etc/ConsolePi/.static.yaml', {})
        self.FALLBACK_USER = self.static.get('FALLBACK_USER', 'pi')
        self.REM_LAUNCH = self.static.get('REM_LAUNCH', '/etc/ConsolePi/src/remote_launcher.py')
        self.cfg_yml = self.get_config_all(yaml_cfg=self.static.get('CONFIG_FILE_YAML'),
                                           legacy_cfg=self.static.get('CONFIG_FILE'))
        self.cfg = self.cfg_yml.get('CONFIG')
        self.ztp = self.cfg_yml.get('ZTP', {})
        self.ovrd = self.cfg_yml.get('OVERRIDES', {})
        self.do_overrides()
        self.picocom_ver = utils.get_picocom_ver()
        self.ser2net_ver = utils.get_ser2net_ver()
        if utils.valid_file("/etc/ser2net.yaml"):
            self.ser2net_file = Path("/etc/ser2net.yaml")
            self.ser2net_conf = self.get_ser2netv4()
        elif utils.valid_file("/etc/ser2net.conf"):
            self.ser2net_file = Path("/etc/ser2net.conf")
            self.ser2net_conf = self.get_ser2netv3()
        else:
            self.ser2net_file = None
            self.ser2net_conf = {}
            log.warning("No ser2net configuration found to extract serial port settings from, using defaults", show=True)

        self.debug = self.cfg.get('debug', False)
        self.cloud = self.cfg.get('cloud', False)
        self.cloud_svc = self.cfg.get('cloud_svc', 'gdrive')
        try:
            self.loc_user = os.getlogin()
        except Exception:
            self.loc_user = os.getenv('SUDO_USER', os.getenv('USER'))

        self.linked_exists = False  # updated in get_outlets_from_file()
        self.hosts = self.get_hosts()
        self.power = self.cfg.get('power', False)
        self.do_dli_menu = None  # updated in get_outlets_from_file()
        self.outlet_groups = {}  # updated in get_outlets_from_file()
        self.outlets = {} if not self.power else self.get_outlets_from_file()
        self.remotes = self.get_remotes_from_file()
        self.remote_update = self.get_remotes_from_file
        self.root = True if os.geteuid() == 0 else False

    def verify_ser2net_file(self):
        """Validate and update ser2net file.

        This is necessary after Instantiation as user could
        convert/remove old .conf file.  Which would cause
        and exception.
        """
        if self.ser2net_file.exists():
            return self.ser2net_file
        elif utils.valid_file("/etc/ser2net.yaml"):
            self.ser2net_file = Path("/etc/ser2net.yaml")
        elif utils.valid_file("/etc/ser2net.conf"):
            self.ser2net_file = Path("/etc/ser2net.conf")
        else:
            self.ser2net_file = None

        return self.ser2net_file

    def get_remotes_from_file(self):
        return self.get_json_file(self.static.get('LOCAL_CLOUD_FILE'))

    def get_config_all(self, yaml_cfg=None, legacy_cfg=None):
        '''Parse bash style cfg vars from cfg file convert to class attributes.'''
        # prefer yaml file for all config items if it exists
        do_legacy = True
        yml = {}
        if yaml_cfg and utils.valid_file(yaml_cfg):
            yml = self.get_yaml_file(yaml_cfg)
            cfg = yml.get('CONFIG', yml)
            if cfg:
                do_legacy = False
                for k in cfg:
                    if cfg[k] in ['true', 'false']:
                        cfg[k] = True if cfg[k] == 'true' else False
                if 'CONSOLEPI_VER' not in yml:
                    yml['CONFIG'] = cfg
        else:
            cfg = {}
            if utils.valid_file(legacy_cfg):
                with open(legacy_cfg, 'r') as config:
                    for line in config:
                        if line.strip() and not line.startswith('#'):
                            var = line.split("=")[0]
                            value = line.split('#')[0]
                            value = value.replace('{0}='.format(var), '')
                            value = value.split('#')[0].strip()

                            if value in ['true', 'false']:
                                cfg[var] = bool(value.replace('false', ''))
                            else:
                                cfg[var] = value.replace('"', '')

        return yml if not do_legacy else {'CONFIG': cfg}

    def do_overrides(self):
        '''Process Globals that can be overriden by values in ConsolePi.yaml
        '''
        ovrd = self.ovrd if self.ovrd else {}
        for k in ovrd:
            if ovrd[k] in ['true', 'false']:
                ovrd[k] = True if ovrd[k] == 'true' else False
        self.ovrd = ovrd
        self.default_baud = ovrd.get('default_baud', DEFAULT_BAUD)
        self.default_dbits = ovrd.get('default_dbits', DEFAULT_DBITS)
        self.default_parity = ovrd.get('default_parity', DEFAULT_PARITY)
        self.default_flow = ovrd.get('default_flow', DEFAULT_FLOW)
        self.default_sbits = ovrd.get('default_sbits', DEFAULT_SBITS)
        self.cloud_pull_only = ovrd.get('cloud_pull_only', False)
        self.compact_mode = ovrd.get('compact_mode', False)
        if ovrd.get("remote_timeout"):
            if isinstance(ovrd["remote_timeout"], dict):  # New config style allowing per host timeout override
                self.remote_timeout = RemoteTimeout()
                for k, v in ovrd["remote_timeout"].items():
                    self.remote_timeout.add_host(k, int(v))
            else:
                self.remote_timeout = RemoteTimeout(int(ovrd["remote_timeout"]))  # Old config style they overrode the default for all remotes
        else:
            self.remote_timeout = RemoteTimeout()  # Default
        self.dli_timeout = int(ovrd.get('dli_timeout', DEFAULT_DLI_TIMEOUT))
        self.so_timeout = int(ovrd.get('smartoutlet_timeout', DEFAULT_SO_TIMEOUT))
        self.cycle_time = int(ovrd.get('cycle_time', DEFAULT_CYCLE_TIME))
        self.api_port = int(ovrd.get("api_port", DEFAULT_API_PORT))
        self.hide_legend = ovrd.get("hide_legend", False)
        # Additional override settings not needed by the python files
        # ovpn_share:  Share VPN connection when wired_dhcp enabled with hotspot connected devices

    def get_outlets_from_file(self):
        '''Get outlets defined in config

        returns:
            dict: with following keys (all values are dicts)
                linked: linked outlets from config (linked to serial adapters- auto pwr-on)
                dli_power: dict any dlis in config have all ports represented here
                failures: failure to connect to any outlets will result in an entry here
                    outlet_name: failure description
        '''
        outlet_data = self.cfg_yml.get('POWER')
        if not outlet_data:  # fallback to legacy json config
            outlet_data = self.get_json_file(self.static.get('POWER_FILE'))

        if not outlet_data:
            if self.power:
                log.show('Power Function Disabled - Configuration Not Found')
                self.power = False
            self.outlet_types = []
            return outlet_data

        types = []
        by_dev: Dict[str, Any] = {}
        for k in outlet_data:
            if k == "OUTLET-GROUPS":
                self.outlet_groups = outlet_data[k]
                continue
            _type = outlet_data[k].get('type').lower()
            relays = [] if _type != "esphome" else utils.listify(outlet_data[k].get('relays', k))
            linked = outlet_data[k].get('linked_devs', {})

            if linked:
                outlet_data[k]['linked_devs'] = utils.format_dev(outlet_data[k]['linked_devs'],
                                                                 hosts=self.hosts, outlet_groups=self.outlet_groups, with_path=True)
                self.linked_exists = True
                for dev in outlet_data[k]['linked_devs']:
                    if _type == 'dli':
                        self.do_dli_menu = True
                        _this = [f"{k}:{[int(p) for p in utils.listify(outlet_data[k]['linked_devs'][dev])]}"]
                    elif _type == 'esphome':
                        _linked = utils.listify(outlet_data[k]['linked_devs'][dev])
                        _this = [f'{k}:{[p for p in _linked]}']
                    else:
                        _this = [k]
                    by_dev[dev] = _this if dev not in by_dev else by_dev[dev] + _this
            else:
                outlet_data[k]['linked_devs'] = {}

            if outlet_data[k]["type"].lower() not in types:
                types.append(outlet_data[k]["type"].lower())

            if outlet_data[k]['type'].upper() == 'GPIO' and isinstance(outlet_data[k].get('address'), str) \
                    and outlet_data[k]['address'].isdigit():
                outlet_data[k]['address'] = int(outlet_data[k]['address'])

            # This block determines if we should show dli_menu / if any esphome outlets match criteria to show
            # in dli menu (anytime it has exactly 8 outlets, if it has > 1 relay and not all are linked)
            if not self.do_dli_menu and _type == "esphome" and len(relays) > 1:
                if len(relays) == 8 or not linked:
                    self.do_dli_menu = True
                elif [r for r in relays if f"'{r}'" not in str(linked)]:
                    self.do_dli_menu = True

        self.outlet_types = types
        outlet_data = {
            'defined': {k: v for k, v in outlet_data.items() if k != "OUTLET-GROUPS"},
            'linked': by_dev,
            'dli_power': {},
            'esp_power': {},
            'failures': {}
        }

        return outlet_data

    def get_json_file(self, json_file):
        '''Return dict from json file.'''
        if os.path.isfile(json_file) and os.stat(json_file).st_size > 0:
            with open(json_file) as f:
                try:
                    return json.load(f)
                except ValueError as e:
                    log.warning(f'Unable to load configuration from {json_file}\n\t{e}', show=True)

    def get_yaml_file(self, yaml_file):
        '''Return dict from yaml file.'''
        if os.path.isfile(yaml_file) and os.stat(yaml_file).st_size > 0:
            with open(yaml_file) as f:
                try:
                    # return yaml.load(f, Loader=yaml.BaseLoader)
                    return yaml.load(f, Loader=yaml.SafeLoader)
                except ValueError as e:
                    log.warning(f'Unable to load configuration from {yaml_file}\n\t{e}', show=True)

    def get_hosts(self):
        '''Parse user defined hosts for inclusion in menu

        returns dict with formatted keys prepending /host/
        '''
        # utils = self.utils
        hosts = self.cfg_yml.get('HOSTS')
        if not hosts:  # fallback to legacy json config
            hosts = self.get_json_file(self.static.get('REM_HOSTS_FILE'))
            if not hosts:
                return {}

        # generate remote command used in menu
        for h in hosts:
            hosts[h]["method"] = hosts[h].get('method', 'ssh').lower()
            if hosts[h]["method"] == 'ssh':  # method defaults to ssh if not provided
                port = 22 if ':' not in hosts[h]['address'] else hosts[h]['address'].split(':')[1]
                _user_str = '' if not hosts[h].get('username') else f'{hosts[h].get("username")}@'
                key_file = None
                if hosts[h].get("key") and self.loc_user is not None:
                    if utils.valid_file(f"/home/{self.loc_user}/.ssh/{hosts[h]['key']}"):
                        user_key = Path(f"/home/{self.loc_user}/.ssh/{hosts[h]['key']}")
                        if utils.valid_file(f"/etc/ConsolePi/.ssh/{hosts[h]['key']}"):
                            mstr_key = Path(f"/etc/ConsolePi/.ssh/{hosts[h]['key']}")
                            if mstr_key.stat().st_mtime > user_key.stat().st_mtime:
                                shutil.copy(mstr_key, user_key)
                                shutil.chown(user_key, user=self.loc_user, group=self.loc_user)
                                user_key.chmod(0o600)
                                log.info(f"{hosts[h]['key']} Updated from ConsolePi global .ssh key_dir to "
                                         f"{str(user_key.parent)} for use with {h}...", show=True)
                        key_file = str(user_key)
                    elif utils.valid_file(hosts[h]['key']):
                        key_file = hosts[h]['key']
                    elif utils.valid_file(f"/etc/ConsolePi/.ssh/{hosts[h]['key']}"):
                        user_ssh_dir = Path(f"/home/{self.loc_user}/.ssh/")
                        if user_ssh_dir.is_dir():
                            shutil.copy(f"/etc/ConsolePi/.ssh/{hosts[h]['key']}", user_ssh_dir)
                            user_key = Path(f"{user_ssh_dir}/{hosts[h]['key']}")
                            shutil.chown(user_key, user=self.loc_user, group=self.loc_user)
                            user_key.chmod(0o600)
                            log.info(f"{hosts[h]['key']} imported from ConsolePi global .ssh key_dir to "
                                     f"{str(user_ssh_dir)} for use with {h}...", show=True)
                            key_file = str(user_key)
                hosts[h]['cmd'] = (
                    f"sudo -u {self.loc_user} ssh{' ' if not key_file else f' -i {key_file} '}" f"-t {_user_str}{hosts[h]['address'].split(':')[0]} -p {port}"
                )
                # hosts[h]['cmd'] = f"sudo -u {self.loc_user} ssh -t {_user_str}{hosts[h]['address'].split(':')[0]} -p {port}"
            elif hosts[h].get('method').lower() == 'telnet':
                port = 23 if ':' not in hosts[h]['address'] else hosts[h]['address'].split(':')[1]
                _user_str = '' if not hosts[h].get('username') else f'-l {hosts[h].get("username")}'
                hosts[h]['cmd'] = f"sudo -u {self.loc_user} telnet {_user_str} {hosts[h]['address'].split(':')[0]} {port}"

        groups = [hosts[h].get('group', 'user-defined') for h in hosts]
        host_dict = {'main': {}, 'rshell': {}}

        for g in utils.unique(groups):
            host_dict['main'][g] = {f'/host/{h.split("/")[-1]}': hosts[h] for h in hosts
                                    if hosts[h].get('show_in_main', False) and hosts[h].get('group', 'user-defined') == g}
            if not host_dict['main'][g]:
                del host_dict['main'][g]
            host_dict['rshell'][g] = {f'/host/{h.split("/")[-1]}': hosts[h] for h in hosts
                                      if not hosts[h].get('show_in_main', False) and hosts[h].get('group', 'user-defined') == g}
            if not host_dict['rshell'][g]:
                del host_dict['rshell'][g]

        host_dict['_methods'] = utils.unique([hosts[h].get('method', 'ssh') for h in hosts])
        host_dict['_host_list'] = [f'/host/{h.split("/")[-1]}' for h in hosts]

        return host_dict

    def get_ser2net(self):
        self.verify_ser2net_file()
        return self.get_ser2netv4() if self.ser2net_file.suffix in [".yaml", ".yml"] else self.get_ser2netv3()

    @staticmethod
    def get_v4_line(tty_dev: str, *, port: int, baud: int, parity: str, dbits: int, sbits: int, flow: str, logfile: str | Path, log_ptr: str, **kwargs) -> str:
        ser2net_flow = {
            'n': '',
            'x': ' XONXOFF',
            'h': ' RTSCTS'
        }
        ser2netv4_line = f"""connection: &{tty_dev.replace("/dev/", "")}
  accepter: telnet(rfc2217),tcp,{port}
  connector: serialdev,{tty_dev},{baud}{parity}{dbits}{sbits},local{'' if flow == 'n' else ',' + ser2net_flow[flow].lower()}
  enable: on
  options:
    banner: *banner
    kickolduser: true
    telnet-brk-on-sync: true
"""
        if logfile and log_ptr:
            log_type = log_ptr.split("=")[0]
            log_alias = "".join(log_ptr.split("=")[1:])
            pointers = {
                "tr": "trace-read",
                "tw": "trace-write",
                "tb": "trace-both",
            }
            ser2netv4_line = f'{ser2netv4_line}    {pointers[log_type]}: *{log_alias}\n'
        return ser2netv4_line

    def get_ser2netv3(self):
        '''Parse ser2net.conf to extract connection info for serial adapters

        retruns 2 level dict (empty dict if ser2net.conf not found or empty):
            {
                <adapter name or alias>: {
                    "baud": <baud>,
                    "dbits": <data bits>,
                    "flow": "<flow control>",
                    "parity": "<parity>",
                    "sbits": <stop bits>,
                    "port": <telnet port (ser2net),
                    "logfile": None or logfile if defined in ser2net.conf
                    "cmd": picocom command string used in menu
                    "line": The line from ser2net.conf
                }
            }
        '''
        ########################################################
        # --- ser2net (3.x) config lines look like this ---
        # ... 9600 NONE 1STOPBIT 8DATABITS XONXOFF LOCAL -RTSCTS
        # ... 9600 8DATABITS NONE 1STOPBIT banner
        ########################################################
        if self.ser2net_ver and int(self.ser2net_ver[0]) == 4:
            log.warning("You are running ser2net v4 which uses /etc/ser2net.yaml for it's config.  However the v3 config /etc/ser2net.conf was found.", show=True)
            log.warning("The menu will continue to use ser2net.conf if present in case ser2net config has not been migrated.", show=True)
            log.warning("/etc/ser2net.yaml will be used to extract serial settings once /etc/ser2net.conf is removed", show=True)

        ser2net_conf = {}
        trace_files = {}
        cfg = self.ser2net_file.read_text()
        for line in cfg.splitlines():
            if line.lstrip().startswith("#"):
                continue
            elif not line.strip():
                continue
            elif 'TRACEFILE:' in line:
                line = line.split(':')
                trace_files[line[1]] = "".join(line[2:])
                continue
            elif not line[0].isdigit():
                continue
            _line = line.strip('\n')
            line = line.split(':')
            tty_port = int(line[0])
            tty_dev = line[3]

            # Reset defaults
            # baud is used to determine parsing failure
            dbits = 8
            parity = 'n'
            flow = 'n'
            sbits = 1
            logfile = None
            log_ptr = None

            connect_params = line[4].replace(',', ' ').split()
            baud = None
            for option in connect_params:
                if option in self.static.get(
                    'VALID_BAUD',
                    ['300', '1200', '2400', '4800', '9600', '19200', '38400', '57600', '115200']
                ):
                    baud = int(option)
                elif 'DATABITS' in option:
                    dbits = int(option.replace('DATABITS', ''))  # int 5 - 8
                    if dbits < 5 or dbits > 8:
                        log.warning(
                            f'{tty_dev}: Invalid value for "data bits" found in ser2net.conf falling back to 8',
                            show=True)
                        dbits = 8
                elif option in ['EVEN', 'ODD', 'NONE']:
                    parity = option[0].lower()  # converts to e o n used by picocom
                elif option == 'XONXOFF':
                    flow = 'x'
                elif option == 'RTSCTS':
                    flow = 'h'
                elif 'STOPBIT' in option:   # Not used by picocom
                    sbits = int(option[0]) if option[0].isdigit else 1
                elif 'tb=' in option or 'tr=' in option or 'tw=' in option:
                    log_ptr = option
                    logfile = option.split('=')[1]

            # Use baud to determine if options were parsed correctly
            if baud is None:
                log.warning(f'{tty_dev} found in ser2net but unable to parse baud falling back to {self.default_baud}',
                            show=True)
                baud = self.default_baud

            # parse TRACEFILE defined in ser2net.conf
            cmd_base = f'picocom {tty_dev} --baud {baud} --flow {flow} --databits {dbits} --parity {parity}'
            if self.picocom_ver > 1:  # picocom ver 1.x in Stretch doesn't support "--stopbits"
                cmd_base = cmd_base + f' --stopbits {sbits}'
            if logfile:
                logfile = trace_files[logfile]
                logfile = logfile.replace('\\p', str(tty_port)).replace('\\d', tty_dev.split('/')[-1])
                logfile = logfile.replace('\\s', f'{baud}_{dbits}{parity.upper()}{sbits}')
                logfile = logfile.split('\\')[0] + '-{{timestamp}}.log'  # + time.strftime('%H.%M.log')
                cmd = cmd_base + f' --logfile {logfile}'
                utils.do_shell_cmd(f"mkdir -p {'/'.join(logfile.split('/')[0:-1])}")
                utils.set_perm('/'.join(logfile.split('/')[0:-1]))
            else:
                cmd = cmd_base

            # update dict with values for this device
            ser2net_conf[tty_dev] = {
                'port': tty_port,
                'baud': baud,
                'dbits': dbits,
                'parity': parity,
                'flow': flow,
                'sbits': sbits,
                'logfile': logfile,
                'log_ptr': log_ptr,
                'cmd': cmd,
                'line': _line,
            }
            ser2net_conf[tty_dev]['v4line'] = self.get_v4_line(tty_dev, **ser2net_conf[tty_dev])
            if trace_files:
                _v4_tracefiles = [f'define: &{k} {v}' for k, v in trace_files.items()]
                ser2net_conf["_v4_tracefiles"] = "# TRACEFILE DEFINITIONS\n" + "\n".join(_v4_tracefiles)

        return ser2net_conf

    def get_ser2netv4(self, file: Path = None) -> Dict:
        """Parse ser2net.yaml (ser2net 4.x) to extract connection info for serial adapters

        Args:
            ser_dict (dict, optional): Used to convert ser2netv3 to v4. Defaults to None.

        retruns 2 level dict (empty dict if ser2net.yaml not found or empty):
            {
                <adapter name or alias>: {
                    "baud": <baud>,
                    "dbits": <data bits>,
                    "flow": "<flow control>",
                    "parity": "<parity>",
                    "sbits": <stop bits>,
                    "port": <telnet port (ser2net),
                    "logfile": None or logfile if defined in ser2net.conf
                    "cmd": picocom command string used in menu
                    "line": The config lines from ser2net.yaml
                }
            }
        """
        ########################################################
        # --- ser2net (4.x) config lines look like this ---
        # connection: &con0096
        #   accepter: telnet(rfc2217),tcp,2000
        #   connector: serialdev,/dev/ttyS0,9600n81,local
        #   enable: on
        #   options:
        #     banner: *banner
        #     kickolduser: true
        #     telnet-brk-on-sync: true
        ########################################################
        ser2net_conf = {}
        if file is None and self.ser2net_file is None:
            return {}

        raw = self.ser2net_file.read_text() if not file else file.read_text()
        raw_mod = "\n".join([line if not line.startswith("connection") else f"{line.split('&')[-1]}:" for line in raw.splitlines()])
        banner = [line for line in raw.splitlines() if line.startswith("define: &banner")]
        banner = None if not banner else banner[-1].replace("define: &banner", "").lstrip()
        try:
            ser_dict = yaml.safe_load(raw_mod)
        except Exception as e:
            log.error(f"Error ({e.__class__.__name__}) occured parsing {file or self.ser2net_file}, verify file contents.", show=True)
            log.exception(e)
            return {}

        for k, v in ser_dict.items():
            if not isinstance(v, dict):
                continue
            if "accepter" not in v or "connector" not in v:
                continue
            if "serialdev" not in v["connector"]:
                log.info(f"skipping ser2net config for {v['connector']} as it is not a serialdev")
                continue

            con_dict = {f"connection: &{k}": v}
            tty_port = v["accepter"].split(",")[-1]
            tty_port = 0 if not tty_port.isdigit() else int(tty_port)
            _connector = v["connector"]
            _connector_list = [part.strip() for part in _connector.split(",")]
            tty_dev = _connector_list[1]
            _baud_str = _connector_list[2].rstrip() if " " not in _connector_list[2].rstrip() else _connector_list[2].split(" ")[0]
            _parity_options = "neoms"  # None Even Odd Mark Space
            parity = [x for x in list(_parity_options) if x in _baud_str]
            if not parity:
                log.warning(f"unable to extract parity from {_baud_str} for {_connector} defaulting to 'n'")
                parity = "n"
            else:
                parity = parity[0]
            baud = _baud_str[0:_baud_str.index(parity)]
            if not baud.isdigit():
                log.warning(f"parse error extracting baud ({baud}) from ser2net.yaml for {_connector}, defaulting to {self.default_baud}", show=True)
                baud = self.default_baud
            else:
                baud = int(baud)

            if parity in ["m", "s"]:
                parity = "n"
                log.warning(f"Setting parity to none for {_connector} when launched via menu. picocom does not support 'm'ark or 's'pace", show=True)
            # parity = _connector_list[2][-3]
            if "rtscts" in str(_connector_list).lower():
                flow = "h"
            elif "xonxoff" in str(_connector_list).lower():  # TODO need to verify this is the format ser2net.yaml accepts, documentation doesn't mention what the options are
                flow = "x"
            else:
                flow = "n"
            dbits = int(_baud_str[-2])
            sbits = int(_baud_str[-1])
            if sbits not in range(1, 3):
                log.warning(f"parse error of config error stop bit ({sbits}) extracted from ser2net.yaml for {_connector}, is invalid defaulting to 1", show=True)
                sbits = 1
            if dbits < 5 or dbits > 8:
                log.warning(
                    f'{tty_dev}: Invalid value for "data bits" found in ser2net.conf falling back to 8',
                    show=True)
                dbits = 8
            if "options" in v:
                keys = [k for k in v["options"].keys() if k.startswith("trace")]
                # picocom just has the simple log option ser2net allows multiple with various options use trace-both if specified
                # it's the closest to what picocom will do.  otherwise simply base it on whatever the first trace entry is.
                if keys:
                    _log_key = "trace-both" if len(keys) > 1 and "trace-both" in keys else keys[0]
                logfile = log_ptr = None if not keys else v["options"][_log_key]

            # parse TRACEFILE defined in ser2net.yaml se we can pass a similar log file format to picocom when launched from menu
            cmd_base = f'picocom {tty_dev} --baud {baud} --flow {flow} --databits {dbits} --parity {parity}'
            if self.picocom_ver > 1:  # picocom ver 1.x in Stretch doesn't support "--stopbits"
                cmd_base = cmd_base + f' --stopbits {sbits}'
            if logfile:
                logfile = logfile.replace('\\p', str(tty_port)).replace('\\d', tty_dev.split('/')[-1])
                logfile = logfile.replace('\\s', f'{baud}_{dbits}{parity.upper()}{sbits}')
                logfile = logfile.split('\\')[0] + '-{{timestamp}}.log'  # + menu will convert totime.strftime('%H.%M.log') when session is launched
                cmd = cmd_base + f' --logfile {logfile}'
                if not Path(logfile).parent.is_dir():
                    Path(logfile).parent.mkdir()
                    utils.set_perm(str(Path(logfile).parent))
            else:
                cmd = cmd_base

            # update dict with values for this device
            ser2net_conf[tty_dev] = {
                'port': tty_port,
                'baud': baud,
                'dbits': dbits,
                'parity': parity,
                'flow': flow,
                'sbits': sbits,
                'logfile': logfile,
                'log_ptr': log_ptr,
                'cmd': cmd,
                'line': f'{list(con_dict.keys())[0]}\n{"".join(line if "banner:" not in line else line.replace(f"banner: {banner}", "banner: *banner") for idx, line in enumerate(yaml.safe_dump(con_dict, indent=2).splitlines(keepends=True)) if idx > 0)}'
            }

        return ser2net_conf



#!/etc/ConsolePi/venv/bin/python3

import sys

sys.path.insert(0, '/etc/ConsolePi/src/pypkg')
from consolepi import config  # type: ignore # NoQA

cfg = {**config.cfg, **{"comment": "START OF OVERRIDES"},  **config.ovrd}


def get_config(cfg=cfg):
    '''get ConsolePi configuration from ConsolePi

    ConsolePi config.py will use ConsolePi.yaml if exists, ConsolePi.json if it doesn't

    Keyword Arguments:
        cfg {str} -- "static" or "bash" static gets static system vars from .static.yaml
                     "bash" or the default will get user config from ConsolePi.yaml or
                     ConsolePi.json (default: {cfg})

    Output is bash formatted var=value file for use as a source in bash scripts
    '''
    for k, v in cfg.items():
        if isinstance(v, dict):
            continue  # currently only applies to remote_timeout
        elif isinstance(v, str):
            print(f"{k}='{v}'")
        elif isinstance(v, list):
            v = ' '.join(v)
            print(f'{k}=({v})')
        elif isinstance(v, bool):
            v = 'true' if v else 'false'
            print(f'{k}={v}')
        elif not v:
            print(f'{k}=')
        else:
            print(f'{k}={v}')


if __name__ == '__main__':
    if len(sys.argv) > 1:
        if sys.argv[1] == 'bash':
            get_config(cfg)
        elif sys.argv[1] == 'static':
            get_config(config.static)
        else:
            print(
                f"\nInvalid Argument {sys.argv[1]}\n    "
                "Valid Args:\n    "
                "'bash': Get main Configuration (Default if no args)\n    "
                "'static': Get static variables\n"
                  )
            sys.exit(1)
    else:
        get_config()



#!/etc/ConsolePi/venv/bin/python3

import os
import yaml
from consolepi import log, config, utils  # type: ignore


class Rename():
    def __init__(self, menu):
        self.menu = menu
        self.udev_pending = False
        self.baud = config.default_baud
        self.data_bits = config.default_dbits
        self.parity = config.default_parity
        self.flow = config.default_flow
        self.sbits = config.default_sbits
        self.parity_pretty = {'o': 'Odd', 'e': 'Even', 'n': 'No'}
        self.flow_pretty = {'x': 'Xon/Xoff', 'h': 'RTS/CTS', 'n': 'No'}
        self.rules_file = config.static.get('RULES_FILE', '/etc/udev/rules.d/10-ConsolePi.rules')
        self.ttyama_rules_file = config.static.get('TTYAMA_RULES_FILE', '/etc/udev/rules.d/11-ConsolePi-ttyama.rules')
        self.ser2net_file = config.static.get('SER2NET_FILE', '/etc/ser2net.conf')  # TODO Not Used Remove once verified all refs switched to config.ser2net_file
        self.reserved_names = ['ttyUSB', 'ttyACM', 'ttyAMA', 'ttySC']

    # --- // START MONSTER RENAME FUNCTION \\ --- # TODO maybe break this up a bit
    def do_rename_adapter(self, from_name):
        '''Rename USB to Serial Adapter

        Creates new or edits existing udev rules and ser2net conf
        for USB to serial adapters detected by the system.

        params:
        from_name(str): Devices current name passed in from rename_menu()

        returns:
        None type if no error, or Error (str) if Error occurred
        '''
        from_name = from_name.replace('/dev/', '')
        local = self.cpi.local
        c = {
            'green': '\033[1;32m',  # Bold with normal ForeGround
            'red': '\033[1;31m',
            'norm': '\033[0m',  # Reset to Normal
        }
        c_from_name = '{}{}{}'.format(c['red'], from_name, c['norm'])
        error = False
        use_def = True

        try:
            to_name = None
            while not to_name:
                print(" Press 'enter' to keep the same name and change baud/parity/...")
                to_name = input(f' [rename {c_from_name}]: Provide desired name: ')
                print("")
                to_name = to_name or from_name
            to_name = to_name.replace('/dev/', '')  # strip /dev/ if they thought they needed to include it
            # it's ok to essentialy rename with same name (to chg baud etc.), but not OK to rename to a name that is already
            # in use by another adapter
            # TODO collect not connected adapters as well to avoid dups
            if from_name != to_name and f"/dev/{to_name}" in local.adapters:
                return f"There is already an adapter using alias {to_name}"

            for _name in self.reserved_names:
                if to_name.startswith(_name):
                    return f"You can't start the alias with {_name}.  Matches system root device prefix"

            if ' ' in to_name or ':' in to_name or '(' in to_name or ')' in to_name:
                print('\033[1;33m!!\033[0m Spaces, Colons and parentheses are not allowed by the associated config files.\n'
                      '\033[1;33m!!\033[0m Swapping with valid characters\n')
                to_name = to_name.replace(' ', '_').replace('(', '_').replace(')', '_')  # not allowed in udev
                to_name = to_name.replace(':', '-')  # replace any colons with - as it's the field delim in ser2net

        except (KeyboardInterrupt, EOFError):
            return 'Rename Aborted based on User Input'

        c_to_name = f'{c["green"]}{to_name}{c["norm"]}'
        log_c_to_name = "".join(["{{green}}", to_name, "{{norm}}"])

        go, con_only = True, False
        if from_name == to_name:
            log.show(f"Keeping {log_c_to_name}. Changing connection settings Only.")
            con_only = True
            use_def = False
        elif utils.user_input_bool(' Please Confirm Rename {} --> {}'.format(c_from_name, c_to_name)) is False:
            go = False

        if go:
            for i in local.adapters:
                if i == f'/dev/{from_name}':
                    break
            _dev = local.adapters[i].get('config')  # type: ignore # dict
            # -- these values are always safe, values set by config.py if not extracted from ser2net.conf
            baud = _dev['baud']
            dbits = _dev['dbits']
            flow = _dev['flow']
            sbits = _dev['sbits']
            parity = _dev['parity']
            word = 'keep existing'
            for _name in self.reserved_names:
                if from_name.startswith(_name):
                    word = 'Use default'

            # -- // Ask user if they want to update connection settings \\ --
            if not con_only:
                use_def = utils.user_input_bool(' {} connection values [{} {}{}1 Flow: {}]'.format(
                    word, baud, dbits, parity.upper(), self.flow_pretty[flow]))

            if not use_def:
                self.con_menu(rename=True, con_dict={'baud': baud, 'data_bits': dbits, 'parity': parity,
                                                     'flow': flow, 'sbits': sbits})
                baud = self.baud
                parity = self.parity
                dbits = self.data_bits
                parity = self.parity
                flow = self.flow
                sbits = self.sbits

            # restore defaults back to class attribute if we flipped them when we called con_menu
            # TODO believe this was an old hack, and can be removed
            if hasattr(self, 'con_dict') and self.con_dict:
                self.baud = self.con_dict['baud']
                self.data_bits = self.con_dict['data_bits']
                self.parity = self.con_dict['parity']
                self.flow = self.con_dict['flow']
                self.sbits = self.con_dict['sbits']
                self.con_dict = None

            if word == 'Use default':  # see above word is set if from_name matches a root_dev pfx
                devs = local.detect_adapters()
                if f'/dev/{from_name}' in devs:
                    _tty = devs[f'/dev/{from_name}']
                    id_prod = _tty.get('id_model_id')
                    id_model = _tty.get('id_model')  # NoQA pylint: disable=unused-variable
                    id_vendorid = _tty.get('id_vendor_id')
                    id_vendor = _tty.get('id_vendor')  # NoQA pylint: disable=unused-variable
                    id_serial = _tty.get('id_serial_short')
                    id_ifnum = _tty.get('id_ifnum')
                    id_path = _tty.get('id_path')  # NoQA pylint: disable=unused-variable
                    lame_devpath = _tty.get('lame_devpath')
                    root_dev = _tty.get('root_dev')  # NoQA pylint: disable=unused-variable
                else:
                    return 'ERROR: Adapter no longer found'

                # -- // ADAPTERS WITH ALL ATTRIBUTES AND GPIO UART (TTYAMA) \\ --
                if id_prod and id_serial and id_vendorid:
                    if id_serial not in devs['_dup_ser']:
                        udev_line = (
                            'ATTRS{{idVendor}}=="{}", ATTRS{{idProduct}}=="{}", ATTRS{{serial}}=="{}", SYMLINK+="{}"'.format(id_vendorid, id_prod, id_serial, to_name)
                        )

                        error = None
                        while not error:
                            error = self.add_to_udev(udev_line, '# END BYSERIAL-DEVS')
                            error = self.do_ser2net_line(from_name=from_name, to_name=to_name, baud=baud, dbits=dbits,
                                                         parity=parity, flow=flow)
                            break

                    # -- // MULTI-PORT ADAPTERS WITH COMMON SERIAL (different ifnums) \\ --
                    else:
                        # SUBSYSTEM=="tty", ATTRS{idVendor}=="0403", ATTRS{idProduct}=="6011", ATTRS{serial}=="FT4XXXXP", GOTO="FTXXXXP"  # NoQA
                        udev_line = (
                            'ATTRS{{idVendor}}=="{0}", ATTRS{{idProduct}}=="{1}", ATTRS{{serial}}=="{2}", GOTO="{2}"'.format(id_vendorid, id_prod, id_serial)
                        )

                        error = None
                        while not error:
                            error = self.add_to_udev(udev_line, '# END BYPORT-POINTERS')
                            # ENV{ID_USB_INTERFACE_NUM}=="00", SYMLINK+="FT4232H_port1", GOTO="END"
                            udev_line = ('ENV{{ID_USB_INTERFACE_NUM}}=="{}", SYMLINK+="{}"'.format(id_ifnum, to_name))
                            error = self.add_to_udev(udev_line, '# END BYPORT-DEVS', label=id_serial)
                            error = self.do_ser2net_line(from_name=from_name, to_name=to_name, baud=baud, dbits=dbits,
                                                         parity=parity, flow=flow)
                            break

                else:
                    if f'/dev/{from_name}' in devs:
                        devname = devs[f'/dev/{from_name}'].get('devname', '')
                        # -- // local ttyAMA adapters \\ --
                        if 'ttyAMA' in devname:
                            udev_line = ('KERNEL=="{}", SYMLINK+="{}"'.format(devname.replace('/dev/', ''), to_name))

                            # Testing simplification not using separate file for ttyAMA
                            error = None
                            while not error:
                                error = self.add_to_udev(udev_line, '# END TTYAMA-DEVS')
                                error = self.do_ser2net_line(from_name=from_name, to_name=to_name, baud=baud, dbits=dbits,
                                                             parity=parity, flow=flow)
                                break
                        else:
                            # -- // LAME ADAPTERS NO SERIAL NUM (map usb port) \\ --
                            log.warning('[ADD ADAPTER] Lame adapter missing key detail: idVendor={}, idProduct={}, serial#={}'.format(  # NoQA
                                        id_vendorid, id_prod, id_serial))
                            print('\n\n This Device Does not present a serial # (LAME!).  So the adapter itself can\'t be '
                                  'uniquely identified.\n There are 2 options for naming this device:')

                            mlines = [
                                '1. Map it to the USB port it\'s plugged in to'
                                '\n\tAnytime a {} {} tty device is plugged into the port it\n\tis currently plugged into it will '
                                'adopt the {} alias'.format(
                                    _tty['id_vendor_from_database'], _tty['id_model_from_database'], to_name),
                                '2. Map it by vedor ({0}) and model ({1}) alone.'
                                '\n\tThis will only work if this is the only {0} {1} adapter you plan to plug in'.format(
                                    _tty['id_vendor_from_database'], _tty['id_model_from_database'])
                                # 'Temporary mapping' \
                                # '\n\tnaming will only persist during this menu session\n'
                            ]
                            print(self.menu.format_subhead(mlines))
                            print('\n b. back (abort rename)\n')
                            valid_ch = {
                                '1': 'by_path',
                                '2': 'by_id'
                            }
                            valid = False
                            ch = ''
                            while not valid:
                                print(' Please Select an option')
                                ch = self.wait_for_input()
                                if ch.lower == 'b':
                                    log.show(f'Rename {from_name} --> {to_name} Aborted')
                                    return
                                elif ch.lower in valid_ch:
                                    valid = True
                                else:
                                    print('invalid choice {} Try Again.'.format(ch.orig))

                            udev_line = None
                            if valid_ch[ch.lower] == 'temp':
                                error = True
                                print('The Temporary rename feature is not yet implemented')
                            elif valid_ch[ch.lower] == 'by_path':
                                udev_line = (
                                    'ATTRS{{idVendor}}=="{0}", ATTRS{{idProduct}}=="{1}", GOTO="{0}_{1}"'.format(  # NoQA
                                        id_vendorid, id_prod), 'ATTRS{{devpath}}=="{}", ENV{{ID_USB_INTERFACE_NUM}}=="{}", '\
                                                               'SYMLINK+="{}"'.format(lame_devpath, id_ifnum, to_name),
                                )
                            elif valid_ch[ch.lower] == 'by_id':
                                udev_line = (
                                    'SUBSYSTEM=="tty", ATTRS{{idVendor}}=="{0}", ATTRS{{idProduct}}=="{1}", GOTO="{0}_{1}"'.format(  # NoQA
                                        id_vendorid, id_prod),
                                    'ENV{{ID_USB_INTERFACE_NUM}}=="{}", SYMLINK+="{}", GOTO="END"'.format(id_ifnum, to_name)  # NoQA
                                )
                            else:
                                error = ['Unable to add udev rule adapter missing details', 'idVendor={}, idProduct={}, serial#={}'.format(  # NoQA
                                    id_vendorid, id_prod, id_serial)]

                            while udev_line:
                                error = self.add_to_udev(udev_line[0], '# END BYPATH-POINTERS')
                                error = self.add_to_udev(udev_line[1], '# END BYPATH-DEVS', label='{}_{}'.format(id_vendorid, id_prod))  # NoQA
                                error = self.do_ser2net_line(from_name=from_name, to_name=to_name, baud=baud, dbits=dbits,
                                                             parity=parity, flow=flow)
                                break
                    else:
                        log.error(f'Device {from_name} No Longer Found', show=True)

            # TODO simplify once ser2net existing verified
            else:   # renaming previously named port.
                # -- // local ttyAMA adapters \\ --
                devname = local.adapters[f'/dev/{from_name}']['udev'].get('devname', '')
                rules_file = self.rules_file if 'ttyAMA' not in devname else self.ttyama_rules_file

                cmd = 'sudo sed -i "s/{0}{3}/{1}{3}/g" {2} && grep -q "{1}{3}" {2} && [ $(grep -c "{0}{3}" {2}) -eq 0 ]'.format(
                    from_name,
                    to_name,
                    rules_file,
                    ''
                )
                error = utils.do_shell_cmd(cmd, shell=True)
                if not error:
                    error = self.do_ser2net_line(from_name=from_name, to_name=to_name, baud=baud, dbits=dbits,
                                                 parity=parity, flow=flow)
                else:
                    return [error.split('\n'), 'Failed to change {} --> {} in {}'.format(from_name, to_name, config.ser2net_file)]

            if error:
                log.error(error, show=True)
            else:
                # Update adapter variables with new_name
                local.adapters[f'/dev/{to_name}'] = local.adapters[f'/dev/{from_name}']
                local.adapters[f'/dev/{to_name}']['config']['port'] = config.ser2net_conf[f'/dev/{to_name}'].get('port', 0)
                local.adapters[f'/dev/{to_name}']['config']['cmd'] = config.ser2net_conf[f'/dev/{to_name}'].get('cmd')
                local.adapters[f'/dev/{to_name}']['config']['line'] = config.ser2net_conf[f'/dev/{to_name}'].get('line')
                local.adapters[f'/dev/{to_name}']['config']['log'] = config.ser2net_conf[f'/dev/{to_name}'].get('log')
                local.adapters[f'/dev/{to_name}']['config']['log_ptr'] = config.ser2net_conf[f'/dev/{to_name}'].get('log_ptr')
                _config_dict = local.adapters[f'/dev/{to_name}']['config']
                if not use_def:  # overwrite con settings if they were changed
                    updates = {
                        'baud': baud,
                        'dbits': dbits,
                        'flow': flow,
                        'parity': parity,
                        'sbits': sbits,
                    }
                    local.adapters[f'/dev/{to_name}']['config'] = {**_config_dict, **updates}

                if from_name != to_name:  # facilitates changing con settings without actually renaming
                    del local.adapters[f'/dev/{from_name}']

                self.udev_pending = True    # toggle for exit function if they exit directly from rename memu

                # update first item in first section of menu_body menu uses it to determine if section is a continuation
                try:
                    self.cur_menu.body_in[0][0] = self.cur_menu.body_in[0][0].replace(from_name, to_name)
                    if self.menu.body_in is not None:  # Can be none when called via rename directly
                        self.menu.body_in[0][0] = self.menu.body_in[0][0].replace(from_name, to_name)
                except Exception as e:
                    log.exception(f"[DEV NOTE menu_body update after rename caused exception.\n{e}", show=False)

        else:
            return 'Aborted based on user input'
    # --- // END MONSTER RENAME METHOD \\ ---

    def do_ser2net_line(self, from_name: str = None, to_name: str = None, baud: int = None,
                        dbits: int = None, parity: str = None,
                        flow: str = None, sbits: int = None):
        '''Process Adapter Configuration Changes in ser2net.conf.

        Keyword Arguments:
            from_name {str} -- The Adapters existing name/alias (default: {None})
            to_name {str} -- The Adapters new name/alias (default: {None})
            baud {int} -- Adapter baud (default: {self.baud})
            dbits {int} -- Adapter databits (default: {self.data_bits})
            parity {str} -- Adapter Parity (default: {self.parity})
            flow {str} -- Adapter flow (default: {self.flow})
            sbits {int} -- Adapter stop bits (default: {self.sbits})

        Returns:
            {str|None} -- Returns error text if an error occurs or None if no issues.
        '''
        # TODO add rename support for ser2net.yaml (v4)
        # don't add the new entry to ser2net if one already exists for the alias
        if from_name != to_name and config.ser2net_conf.get(f"/dev/{to_name}"):
            log.info(f"ser2net: {to_name} already mapped to port {config.ser2net_conf[f'/dev/{to_name}'].get('port')}", show=True)
            return

        ser2net_parity = {
            'n': 'NONE',
            'e': 'EVEN',
            'o': 'ODD'
        }
        ser2net_flow = {
            'n': '',
            'x': ' XONXOFF',
            'h': ' RTSCTS'
        }
        baud = self.baud if not baud else baud
        dbits = self.data_bits if not dbits else dbits
        parity = self.parity if not parity else parity
        flow = self.flow if not flow else flow
        sbits = self.sbits if not sbits else sbits
        log_ptr = ''

        cur_line = config.ser2net_conf.get(f'/dev/{from_name}', {}).get('line')
        # TODO shouldn't  this include /dev/ttyAMA ??? the on-board ttl?
        if cur_line and '/dev/ttyUSB' not in cur_line and '/dev/ttyACM' not in cur_line and '/dev/ttySC' not in cur_line:
            new_entry = False
            if config.ser2net_file.suffix in [".yaml", ".yml"]:
                next_port = int(yaml.safe_load(cur_line.replace(": *", ": REF_"))["connection"].get("accepter").split(",")[-1])
            else:
                next_port = cur_line.split(':')[0]  # Renaming existing
            log_ptr = config.ser2net_conf[f'/dev/{from_name}'].get('log_ptr')
            if not log_ptr:
                log_ptr = ''
        else:
            new_entry = True
            if utils.valid_file(config.ser2net_file):
                ports = [v['port'] for k, v in config.ser2net_conf.items() if not k.startswith("_") and 7000 < v.get('port', 0) <= 7999]
                next_port = 7001 if not ports else int(max(ports)) + 1
            else:
                next_port = 7001
                error = utils.do_shell_cmd(f'sudo cp {config.ser2net_file} /etc/', handle_errors=False)
                if error:
                    log.error(f'Rename Menu Error while attempting to cp ser2net.conf from src {error}', show=True)
                    return error  # error added to display in calling method

        if config.ser2net_file.suffix in [".yaml", ".yml"]:
            ser2net_line = f"""connection: &{to_name}
  accepter: telnet(rfc2217),tcp,{next_port}
  connector: serialdev,/dev/{to_name},{baud}{parity}{dbits}{sbits},local{'' if flow == 'n' else ',' + ser2net_flow[flow].lower()}
  enable: on
  options:
    banner: *banner
    kickolduser: true
    telnet-brk-on-sync: true
"""
            # if flow != "n":
            #     ser2net_line = ser2net_line.rstrip("\n")
            #     ser2net_line = f'{ser2net_line},\n            {ser2net_flow[flow].lower()}\n'
        else:
            ser2net_line = (
                '{telnet_port}:telnet:0:/dev/{alias}:{baud} {dbits}DATABITS {parity} {sbits}STOPBIT {flow} banner {log_ptr}'.format(
                    telnet_port=next_port,
                    alias=to_name,
                    baud=baud,
                    dbits=dbits,
                    sbits=sbits,
                    parity=ser2net_parity[parity],
                    flow=ser2net_flow[flow],
                    log_ptr=log_ptr
                )
            )

        # -- // Append to ser2net config \\ --
        if new_entry:
            error = utils.append_to_file(config.ser2net_file, ser2net_line)
        # -- // Rename Existing Definition in ser2net.conf \\ --
        # -- for devices with existing definitions cur_line is the existing line
        elif config.ser2net_file.suffix in [".yaml", ".yml"]:
            # sudo sed -i 's|^connection: &delme1$|connection: \&test1|g'  /etc/ser2net.yaml
            # sudo sed -i 's|^  connector: serialdev,/dev/orange1|  connector: serialdev,/dev/delme1|g'  /etc/ser2net.yaml
            if not cur_line:
                return f'cur_line has no value.  Leaving {config.ser2net_file} unchanged.'
            connection_line = cur_line.splitlines()[0]
            connector_line = [line for line in cur_line.splitlines() if "connector:" in line and f'/dev/{from_name},' in line.replace("\\/", "/")]
            if connector_line:
                if len(connector_line) > 1:
                    error = f'Found {len(connector_line)} lines in {config.ser2net_file} with /dev/{from_name} defined.  Expected 1.  Aborting'
                elif "connection:" not in connection_line:
                    error = f'Unexpected connection line {connection_line} extracted for /dev/{from_name} from {config.ser2net_file}'
                else:
                    new_connection_line = connection_line.replace(f'&{from_name}', f'\&{to_name}')  # NoQA
                    connector_line = connector_line[0]
                    new_connector_line = connector_line.replace(f"/dev/{from_name},", f"/dev/{to_name},")
                    expected_complete_connector_line = f"  connector: serialdev,/dev/{to_name},{baud}{parity}{dbits}{sbits},local{'' if flow == 'n' else ',' + ser2net_flow[flow].lower()}"
                    if new_connector_line != expected_complete_connector_line:
                        if "local" in new_connector_line:
                            new_connector_line = expected_complete_connector_line
                        else:
                            return f'rename requires all serial settings be on the same line i.e. {expected_complete_connector_line}.'
                    cmds = [
                        f"sudo sed -i 's|^{connection_line}$|{new_connection_line}|g'  {config.ser2net_file}",
                        f"sudo sed -i 's|^{connector_line}$|{new_connector_line}|g'  {config.ser2net_file}"
                    ]
                    for cmd in cmds:
                        error = utils.do_shell_cmd(cmd, shell=True)
                        if error:
                            break
        else:
            ser2net_line = ser2net_line.strip().replace('/', r'\/')
            cur_line = cur_line.replace('/', r'\/')
            cmd = "sudo sed -i 's/^{}$/{}/'  {}".format(
                cur_line, ser2net_line, config.ser2net_file
            )
            error = utils.do_shell_cmd(cmd, shell=True)

        if not error:
            config.ser2net_conf = config.get_ser2net()
        else:
            return error

    def add_to_udev(self, udev_line: str, section_marker: str, label: str = None):
        '''Add or edit udev rules file with new symlink after adapter rename.

        Arguments:
            udev_line {str} -- The properly formatted udev line being added to the file
            section_marker {str} -- Match text used to determine where to place the line


        Keyword Arguments:
            label {str} -- The rules file GOTO label used in some scenarios
                           (i.e. multi-port 1 serial) (default: {None})

        Returns:
            {str|None} -- Returns error string if an error occurs
        '''
        found = ser_label_exists = get_next = update_file = alias_found = False  # NoQA
        # alias_name = udev_line.split("SYMLINK+=")[1].split(",")[0].replace('"', "")  # TODO WIP add log when alias is already associated with diff serial
        goto = line = cmd = ''  # init
        rules_file = self.rules_file  # if 'ttyAMA' not in udev_line else self.ttyama_rules_file  Testing 1 rules file
        if utils.valid_file(rules_file):
            with open(rules_file) as x:
                for line in x:
                    # temporary for those that have the original file
                    if 'ID_SERIAL' in line and 'IMPORT' not in line:
                        _old = 'ENV{ID_SERIAL}=="", GOTO="BYPATH-POINTERS"'
                        _new = 'ENV{ID_SERIAL_SHORT}=="", IMPORT{builtin}="path_id", GOTO="BYPATH-POINTERS"'
                        cmd = "sudo sed -i 's/{}/{}/' {}".format(_old, _new, rules_file)
                        update_file = True

                    # No longer including SUBSYSTEM in formatted udev line, redundant given logic @ top of rules file
                    if line.replace('SUBSYSTEM=="tty", ', '').strip() == udev_line.strip():
                        return  # Line is already in file Nothing to do.
                    # elif f'SYMLINK+="{alias_name}",' in line:
                    #     alias_found = True  # NoQA
                        # log.warning(f'{rules_file.split("/")[-1]} already contains an {alias_name} alias\nExisting alias {udev_line}', show=True)
                    if get_next:
                        goto = line
                        get_next = False
                    if section_marker.replace(' END', '') in line:
                        get_next = True
                    elif section_marker in line:
                        found = True
                    elif label and 'LABEL="{}"'.format(label) in line:
                        ser_label_exists = True

                last_line = line
            if update_file:
                error = utils.do_shell_cmd(cmd)
                if error:
                    log.show(error)

            goto = goto.split('GOTO=')[1].replace('"', '').strip() if 'GOTO=' in goto else None
            if goto is None:
                goto = last_line.strip().replace('LABEL=', '').replace('"', '') if 'LABEL=' in last_line else None
        else:
            error = utils.do_shell_cmd(f'sudo cp /etc/ConsolePi/src/{os.path.basename(rules_file)} /etc/udev/rules.d/')
            # TODO switch to pathlib.Path('path...').copy(src, dst)
            found = True
            goto = 'END'

        if goto and 'GOTO=' not in udev_line:
            udev_line = '{}, GOTO="{}"'.format(udev_line, goto)

        if label and not ser_label_exists:
            udev_line = 'LABEL="{}"\\n{}'.format(label, udev_line)

        # -- // UPDATE RULES FILE WITH FORMATTED LINE \\ --
        if found:
            udev_line = '{}\\n{}'.format(udev_line, section_marker)
            cmd = "sudo sed -i 's/{}/{}/' {}".format(section_marker, udev_line, rules_file)
            error = utils.do_shell_cmd(cmd, handle_errors=False)
            if error:
                return error
        else:  # Not Using new 10-ConsolePi.rules template just append to file
            if section_marker == '# END BYSERIAL-DEVS':
                return utils.append_to_file(rules_file, udev_line)
            else:  # if not by serial device the new template is required
                return 'Unable to Add Line, please use the new 10.ConsolePi.rules found in src dir and\n' \
                    'add you\'re current rules to the BYSERIAL-DEVS section.'

    def trigger_udev(self):
        '''reload/trigger udev (udevadm)

        Returns:
            No return unless there is an error.
            Returns {str} if an error occurs.
        '''
        cmd = 'sudo udevadm control --reload && sudo udevadm trigger && systemctl is-active ser2net >/dev/null 2>&1'\
              '&& sudo systemctl stop ser2net && sleep 1 && sudo systemctl start ser2net'
        error = utils.spinner('Triggering reload of udev & ser2net due to name change',
                              utils.do_shell_cmd, cmd, shell=True)
        if not error:
            self.udev_pending = False
        else:
            log.show('Failed to reload udev rules, you may need to rectify manually for adapter names to display correctly')
            log.show(f'Check /var/log/syslog for errors, the rules file ({self.rules_file}) and reattempt {cmd} manually')
            log.show(error)



#!/usr/bin/env python3

import json
import logging
import logging.handlers
import time
from time import sleep
import socket

import requests
from dlipower import PowerSwitch
from requests.auth import HTTPDigestAuth
from requests.models import ChunkedEncodingError

DLI_TIMEOUT = 7
SEQUENCE_DELAY = 1
DEBUG = False
TIMING = False


class Dli_Logger:

    def __init__(self, debug=DEBUG):
        self.debug = debug
        self.log = self.set_log()

    def set_log(self):
        log = logging.getLogger(__name__)
        log.setLevel(logging.INFO if not self.debug else logging.DEBUG)
        handler = logging.handlers.SysLogHandler(address='/dev/log')
        handler.setLevel(logging.INFO if not self.debug else logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s [%(levelname)s] %(message)s')
        handler.setFormatter(formatter)
        log.addHandler(handler)
        return log


class DLI:

    def __init__(self, fqdn: str, username: str = 'admin', password: str = 'admin',
                 use_https: bool = False, timeout: int = DLI_TIMEOUT, log=Dli_Logger().log):
        self.timeout = timeout
        self.log = log
        self.scheme = 'http://' if not use_https else 'https://'
        self.reachable, self.ip = self.check_reachable(fqdn, 443 if self.scheme.split(':')[0] == 'https' else 80)
        self.fqdn = fqdn
        self.base_url = self.scheme + str(self.ip)
        self.outlet_url = self.base_url + '/restapi/relay/outlets/'
        self.username = username
        self.password = password
        self.rest = None
        if self.reachable:
            try:
                self.dli = self.get_session(username, password)
                self.outlets = self.get_dli_outlets()
            except ConnectionError:
                log.warning(f"DLI @ {self.ip} ware reachable, but an exception occured while trying to establish a session")
                self.dli = self.outlets = {}
        else:
            self.dli = self.outlets = {}
        self.pretty = {
            True: 'ON',
            False: 'OFF'
        }
        if TIMING:
            self.hit = 0  # TIMING
            self._hit = 0  # TIMING

    def __len__(self):
        """
        :return: Number of outlets
        """
        return len(self.outlets)

    def __repr__(self):
        """
        display the representation
        """
        if not self.outlets:
            return "Digital Loggers Web Powerswitch " \
                   "{} (UNCONNECTED)".format(self.fqdn)
        output = 'DLIPowerSwitch at {}\n' \
                 'Outlet\t{:<15}\tState\n'.format(self.fqdn, 'Name')
        for port in self.outlets:
            output += '{}\t{:<15}\t{}\n'.format(port, self.outlets[port]['name'],
                                                self.pretty[self.outlets[port]['state']])
        return output

    def __getitem__(self, index):
        if TIMING:
            self._hit += 1
            print('[__getitem__] hit {} processing port {}'.format(self._hit, index))
        outlets = self.get_dli_outlets()  # self.outlets # if self.hit == 1 else self.get_dli_outlets()
        if outlets:
            if isinstance(index, slice):
                ret_val = {}
                for o in outlets:
                    if o >= index.start and o <= index.stop:
                        ret_val[o] = {'state': self.state(o), 'name': outlets[o]['name']}
            elif isinstance(index, list):
                ret_val = {}
                for o in index:
                    ret_val[o] = {'state': self.state(o), 'name': outlets[o]['name']}
            else:
                ret_val = {index: {'state': self.state(index), 'name': outlets[index]['name']}}
        else:
            ret_val = outlets
        if TIMING:
            print('\t{}'.format(ret_val))

        return ret_val

    def check_reachable(self, host, port, timeout=2):
        # if url is passed check dns first otherwise dns resolution failure causes longer delay
        # determine if host is resolvable
        try:
            s = socket.getaddrinfo(host, port, 0, socket.SOCK_STREAM)
            _ip = s[0][4][0]
        except (socket.gaierror, socket.timeout, TimeoutError) as e:
            self.log.error('[DLI] {} is not reachable\n{}'.format(host, e))
            _ip = None
            return False, _ip

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(timeout)
        try:
            sock.connect((_ip, port))
            reachable = True
        except (socket.error, TimeoutError):
            reachable = False
        sock.close()

        return reachable, _ip

    def get_session(self, username: str, password: str, fqdn: str = None):
        '''Get or Renew a session with the dli from requests module.'''
        log = self.log
        fqdn = self.fqdn if fqdn is None else fqdn

        headers = {'Accept': 'application/json', 'Connection': 'keep-alive'}
        f_url = self.base_url + '/restapi/relay/version/'
        if TIMING:
            start = time.time()  # TIMING
        _session = requests.session()
        _session.auth = HTTPDigestAuth(username, password)
        _session.headers = headers
        r = _session.get(f_url, headers=headers)
        if TIMING:
            print('[TIMING] check api ({}): return: {}, {}'.format(fqdn, r.status_code, time.time() - start))  # type: ignore
        if r.headers['Content-Type'] != 'application/json':  # determine if old screen-scrape method is required for older dlis
            log.debug("[DLI] Using webui scraping method for {}, it doesn't appear to support the new rest API".format(fqdn))
            self.rest = False
            try:
                switch = PowerSwitch(hostname=self.ip, userid=username, password=password, timeout=self.timeout)
            except (ConnectionError, ConnectionResetError, ChunkedEncodingError) as e:
                log.error(f"Exception Connecting to {fqdn}. {e}")
                self.dli = self.outlets = {}
                return  # TODO verify calling method handles None
            return switch
        else:   # web power switch pro - use rest API
            log.debug("[DLI] Using rest API method for {}".format(fqdn))
            self.rest = True
            return _session

    def rename(self, port, new_name):
        """Rename the outlet.

        :param port: The outlet to rename (dli outlet numbering starting with 1)
        :param new_name: New name for the outlet
        :returns: True for success, False for Fail
        """
        log = self.log
        if TIMING:
            start = time.time()  # TIMING
        if self.rest:
            headers = {'Content-Type': 'application/json', 'X-Requested-With': 'XMLHttpRequest'}
            data = json.dumps(new_name)
            r = self.dli.put('{}{}/name/'.format(self.outlet_url, port - 1), data=data, headers=headers)
            if r.status_code != 204:
                log.warning('[DLI] {} returned error response to rename request {},{},{}'.format(self.fqdn, r.status_code,
                                                                                                 r.content, r.reason))
            ret = True if r.status_code == 204 else False
        else:
            ret = self.dli.set_outlet_name(outlet=port, name=new_name)
        if TIMING:
            print('[TIMING] {} rename {}: {}'.format(self.fqdn, port, time.time() - start))  # type: ignore
        if ret:
            self.outlets[port]['name'] = self.name(port)
            return self.outlets[port]['name'] == new_name
        else:
            return ret

    def get_dli_outlets(self):
        '''Get Outlet details from dli.

        Uses self.dli session from __init__
        returns: dict of outlets
        {
            port:int {
                'name': name:str,
                'state': state:bool, (True = On)
                }
        }
        '''
        log = self.log
        if TIMING:
            self.hit += 1
            print('[GET OUTLETS] hit {}'.format(self.hit))
            start = time.time()  # TIMING
        outlet_dict = {}
        outlet_list = []
        if self.rest:
            # New dli API takes about 5 seconds to retrieve the outlet data
            timeout = self.timeout + 3 if self.timeout < 6 else self.timeout
            try:
                r = self.dli.get(self.outlet_url, timeout=timeout)
                # outlet_list = json.loads(r.content.decode('UTF-8'))
                outlet_list = r.json()
            except (socket.error, TimeoutError):
                self.reachable = False
                self.dli = self.outlets = {}
            idx = 1
            if self.reachable:
                for outlet in outlet_list:
                    # for _ in ["critical", "cycle_delay", "locked", "physical_state", "transient_state"]:
                    #     # nul = outlet.pop(_)
                    #     del outlet[_]
                    if not (isinstance(outlet, str) and outlet == "error") and not (isinstance(outlet, dict) and outlet.get('error')):
                        outlet_dict[idx] = {'name': outlet['name'], 'state': outlet['state']}
                        idx += 1
                    else:
                        log.error(f"dli returned error {outlet_list['error']}")
                        # TODO return class with error for menu
        else:
            self.reachable = self.check_reachable(self.fqdn, port=443 if 'https' in self.scheme else 80)[0]
            if self.reachable:
                # retry = 0
                for retry in range(0, 1):
                    try:
                        outlet_list = self.dli.statuslist()
                        if outlet_list is None:  # indicates session has timed out.
                            self.verify_legacy()
                            outlet_list = self.dli.statuslist()
                    except AttributeError as e:
                        log.error(f'dlirest.py, get_dli_outlets exceptions occurred attempting to get outlet_list: {e}')
                        continue
                    if outlet_list:  # can be None if dli suffers transient issue
                        for outlet in outlet_list:
                            outlet_dict[outlet[0]] = {'name': outlet[1], 'state': True if outlet[2].upper() == 'ON' else False}
                        break
                    # retry += 1
                if not outlet_list:
                    log.error(f'[DLI GET OUTLETS] dli @ {self.fqdn} reachable, but failed to fetch statuslist (outlet_list)')
                    self.reachable = False
                    self.dli = self.outlets = {}  # TODO maybe update outlets in defined... see error in exec
            else:
                self.reachable = False
                self.dli = self.outlets = {}
        if TIMING:
            print('[TIMING] {} get_dli_outlets: {}'.format(self.fqdn, time.time() - start))  # type: ignore
        return outlet_dict

    def operate_port(self, port, toState=None, func='toggle'):
        '''Toggle or cycle Power on all or a specified port.

        parameters:
            port: The Interface to toggle
        '''
        log = self.log
        bool_state = {
            'ON': True,
            'OFF': False
        }

        if not self.rest:
            self.verify_legacy()

        # --// SUB Toggle or Cycle Power On Specified Port \\--
        def toggle_sub(port, toState):
            '''Toggle or Cycle Power on Outlet.'''
            port_idx = port - 1
            base_url = self.outlet_url + str(port_idx)
            f_url = base_url + '/state/' if func == 'toggle' else '{}/{}/'.format(base_url, func.lower())
            # use rest API available on newer web power switches (i.e. web power switch Pro)
            if self.rest:
                if func == 'toggle':
                    req = getattr(self.dli, 'put')
                    headers = {'Content-Type': 'application/json', 'X-Requested-With': 'XMLHttpRequest'}
                    data = json.dumps(toState)
                elif func == 'cycle':
                    req = getattr(self.dli, 'post')
                    headers = {'X-Requested-With': 'XMLHttpRequest'}
                    data = None
                try:
                    r = req(f_url, data=data, headers=headers, timeout=10)
                except (Exception, OSError) as e:
                    print(e)
                    log.error(f'EXCEPTION: Unable to Connect {base_url} to {func} port:\n\t{e}')
                    return 404  # TODO return a return class can't return meaningful error here as is
                if len(r.text) == 0:
                    return r.status_code
                else:
                    return r.json()  # rest api returns content false with status 200 if state was off when cycle was issued
            else:   # dlipower.PowerSwitch - screen scrape library
                if func == 'toggle':
                    if toState:
                        r = self.dli.on(port)
                    else:
                        r = self.dli.off(port)

                    if not r:  # dlipower.PowerSwitch returns False if operation Success
                        return toState
                    else:
                        # TODO need to get_session and retry, or put logic in to verify prior to ops
                        return '{} Port {} dlipower library gave an unexpected response: {}'.format(
                            self.fqdn, port, r)
                elif func == 'cycle':
                    if curState:
                        if TIMING:
                            start = time.time()
                        r = self.dli.cycle(port)
                        if TIMING:
                            print('[TIMING] {} cycle {}: {}'.format(self.fqdn, port, time.time() - start))  # type: ignore
                        return not r
                    else:
                        return False  # a False response from cycle indicates port was already off nothing occurred
            # -- END TOGGLE SUB --

        # --// Determine what the new powered state should be \\--
        if toState is not None and isinstance(toState, str):        # TODO should be able to remove, refactored to bool
            if toState.lower() in ['on', 'off']:
                toState = True if toState.lower() == 'on' else False
            else:
                log.error('[DLI] invalid toState Passed to function')

            # --// Validate port passed into method \\--
            if isinstance(port, int) and port <= len(self.outlets):
                pass  # valid
            elif isinstance(port, str) and port.lower() == 'all':
                pass  # valid
            elif isinstance(port, list):
                for p in port:
                    if p not in self.outlets.keys():
                        log.error('[DLI] port {} provided in port list: {} is not valid')
            else:
                log.error('[DLI] Invalid Value provided for port {}'.format(port))

        elif toState is None:   # No toState provided set toState based on opposite of curState
            if isinstance(port, int):
                curState = self.state(port)
                toState = not curState
            elif isinstance(port, str):
                if port == 'all':
                    if func == 'toggle':
                        raise Exception('all specified without desired end state')
            else:
                raise Exception('desired state required when port type is not int')

        # --// perform the func on the port(s) \\--#
        # -- single port passed into method --
        if TIMING:
            start = time.time()
        ret_val = []
        if isinstance(port, int):
            ret_val = toggle_sub(port, toState)
            if TIMING:
                print('[TIMING {}] {} {} {}: {}'.format('rest' if self.rest else 'webui',
                                                        self.fqdn, func, port, time.time() - start))  # type: ignore
        # -- keyword 'all' passed into method --
        elif (isinstance(port, str) and port == 'all'):
            if func.lower() == 'toggle':
                toState = 'ON' if toState else 'OFF'
                if self.rest:
                    url = self.base_url + '/outlet?a=' + toState
                    ret_val = self.verify_session(url)
                else:
                    r = self.dli.geturl(url='outlet?a=' + toState)
                    ret_val = 200 if r is not None else 400
            elif func.lower() == 'cycle':
                if self.rest:
                    url = '{}/outlet?a=CCL'.format(self.base_url)
                    ret_val = self.verify_session(url)
                else:
                    r = self.dli.geturl(url='outlet?a=CCL')
                    ret_val = 200 if r is not None else 400
                return ret_val
            else:
                print('Invalid value for func argument')  # TODO logging exception
            if TIMING:
                print('[TIMING] {} {} {}: {}'.format(self.fqdn, func, port, time.time() - start))  # type: ignore
        # -- something else (should be list or slice) passed in --
        else:
            ret_val = []
            for p in port:
                if TIMING:
                    start = time.time()
                r = toggle_sub(p, toState)
                if TIMING:
                    print('[TIMING {}] {} {} {}: {}'.format('rest' if self.rest else 'webui',
                                                            self.fqdn, func, p, time.time() - start))  # type: ignore
                if isinstance(r, tuple):
                    ret_val.append(r[0])
                else:
                    ret_val.append(r)
                sleep(SEQUENCE_DELAY)
            log.debug('DLI {}] Return values for ports: {} = {}'.format(func, port, ret_val))
            ret_val = list(dict.fromkeys(ret_val))  # get rid of all duplicates
            if len(ret_val) == 1:
                ret_val = ret_val[0]
            else:
                for status_code in ret_val:
                    if status_code > 204:
                        ret_val = status_code

        if isinstance(ret_val, bool):
            return ret_val
        elif isinstance(ret_val, int) and ret_val <= 204:    # toggle power
            return toState if isinstance(toState, bool) else bool_state[toState]
        else:
            return 'An Error occurred {}'.format(ret_val)

    def verify_legacy(self):
        '''Verify session is not expired for non-rest dli

        For DLI lpc 7 and prior which uses the dlipower library (screen scrape), this will check if the original session
        is expired and create a new one if it is.  It's called by the 2 methods that perform actions against the outlets
        (operate_port, get_port_info)
        '''
        if not self.dli.statuslist():
            self.dli = self.get_session(self.username,
                                        self.password,
                                        fqdn=self.fqdn)
            self.log.info(f"Session with {self.fqdn} was expired. Renewed Session")
            # TODO validate and log if still error

    def verify_session(self, url: str):
        '''perform http get operation against dli if the response indicates the session is expired get a new session and retry.

        Used by new rest capable dli web power switches (only), for operations against "all" outlets, given no API method
        is available for all operations.
        '''
        log = self.log
        retry = 0
        ret_val = 400
        r = None
        while retry < 3:
            # -- attempt to perform the action --
            r = self.dli.get(url)
            # -- check to see if session expired --
            if r.content.decode('UTF-8').split('URL=')[1].split('"')[0] != '/index.htm':
                log.debug('[DLI VRFY SESSION] Session appears expired for {}. Renewing... {}'.format(
                          self.fqdn, ' Retry ' + str(retry) if retry > 0 else ''))
                self.dli = self.get_session(self.dli.auth.username, self.dli.auth.password, fqdn=self.fqdn)
            else:
                ret_val = r.status_code
                if ret_val != 200:
                    log.error('[DLI VRFY SESSION] call to ' + url + 'returned ' + str(ret_val))
                    print('[DLI VRFY SESSION] call to ' + url + 'returned ' + str(ret_val))
                break
            retry += 1

        if r and r.content.decode('UTF-8').split('URL=')[1].split('"')[0] != '/index.htm':
            log.warn('[DLI VRFY SESSION] Unable to Renew Session for {}'.format(self.fqdn))
            ret_val = 400

        return ret_val

    def get_port_info(self, port: int, fetch: str = 'state'):
        '''
        returns Bool Representing current port state ~ True = ON
        '''
        log = self.log
        _return = None
        if self.outlets is not None:
            if isinstance(port, int) and port <= len(self.outlets):
                if self.rest:
                    _url = self.outlet_url + str(port - 1) + '/{}/'.format(fetch)
                    try:
                        r = self.dli.get(_url, timeout=self.timeout)
                        if r.status_code == 200:
                            _return = r.json()  # TODO - error exception catch
                        else:
                            _msg = f'[DLI] Bad status code {r.status_code} retunred while checking current {fetch} of port'
                            log.error(_msg)
                            _return = _msg
                    except (socket.error, TimeoutError):
                        self.reachable = False
                        log.error('[DLI] {} appears to be unreachable now'.format(_url))
                        _return = 'Error: [DLI] {} appears to be unreachable now'.format(_url)
                else:
                    self.verify_legacy()
                    if fetch == 'state':
                        _return = self.dli.status(port)
                        if _return.upper() == 'ON':
                            _return = True
                        elif _return.upper() == 'OFF':
                            _return = False
                        else:
                            log.error('[DLI] {} returned invalid state "{}" for port {}'.format(self.fqdn, _return, port))
                    elif fetch == 'name':
                        _return = self.dli.get_outlet_name(port)
            else:
                _return = 'error: invalid port type {}'.format(type(port))
        else:
            _return = 'Error: UNREACHABLE'
        if fetch == 'state' and _return in [True, False]:
            self.outlets[port]['state'] = _return   # ensure outlet dict has current state
        elif fetch == 'name':
            self.outlets[port]['name'] = _return

        return _return

    def toggle(self, port, toState=None):
        return self.operate_port(port, toState=toState)

    def cycle(self, port):
        return self.operate_port(port, toState=None, func='cycle')

    def state(self, port):
        return self.get_port_info(port)

    def name(self, port):
        return self.get_port_info(port, fetch='name')

    def close(self):
        '''
        Be a good citizen and close the session with the dli
        This will help prevent session exaustion on the dli
        This should be called after operations are complete
        '''
        log = self.log
        if self.rest:
            r = self.dli.get(url='{}/logout'.format(self.base_url))
            self.dli.close()
        else:
            r = self.dli.session.get(self.dli.base_url + '/logout')
            self.dli.session.close()
        if r.status_code != 200:
            log.warning('[DLI] Attempt to logout of {0} returned error \n\t{1} {2}: {3}\n\tHeaders: {4}'.format(
                self.fqdn, r.status_code, r.reason, r.text, r.headers))


if __name__ == '__main__':
    pass



#!/etc/ConsolePi/venv/bin/python3

# from consolepi import Response
from consolepi import utils
from consolepi import config
from consolepi.remotes import Remotes
from consolepi.local import Local
from consolepi.power import Outlets
from consolepi.exec import ConsolePiExec
from consolepi.menu import Menu  # NoQA


class ConsolePi():
    def __init__(self, bypass_remotes: bool = False, bypass_outlets: bool = False, bypass_cloud: bool = False):
        self.menu = Menu("main_menu")
        self.local = Local()
        if not bypass_outlets and config.cfg.get('power'):
            self.pwr_init_complete = False
            self.pwr = Outlets()
        else:
            self.pwr_init_complete = True
            self.pwr = None
        self.cpiexec = ConsolePiExec(config, self.pwr, self.local, self.menu)
        if not bypass_remotes:
            self.remotes = Remotes(self.local, self.cpiexec, bypass_cloud=bypass_cloud)

        # TODO Move to menu launch and prompt user
        # verify TELNET is installed and install if not if hosts of type TELNET are defined.
        if config.hosts:
            utils.verify_telnet_installed(config.hosts)



#!/etc/ConsolePi/venv/bin/python3

import json
import sys
# import threading
sys.path.insert(0, '/etc/ConsolePi/src/pypkg')

from consolepi import log # NoQA
from consolepi.consolepi import ConsolePi # NoQA
# from consolepi import threading  # NoQA
# from consolepi.power.outlets import Outlets  # NoQA

cpi = ConsolePi(bypass_remotes=True)
# cpi = ConsolePi()
pwr = cpi.pwr
# utils = cpi.utils
# threading.Thread(target=utils.set_perm, args=['/home/pi/testfile'], name='common_import_test').start()
# outlet_data = cpi.config.outlets
print(pwr.do_esphome_cmd("outlet1.kabrew.com", "outlet1", 'cycle'))
sys.exit(0)

pwr.pwr_start_update_threads()
if cpi.cpiexec.wait_for_threads():
    print('Threads Timed Out')
    sys.exit(1)
else:
    outlets = pwr.data
    print(pwr.do_esphome_cmd("outlet1.kabrew.com", "outlet1", True))

if len(sys.argv) >= 2:
    if len(sys.argv) == 2:
        print(json.dumps(getattr(pwr, sys.argv[1]), indent=4, sort_keys=True))
    else:
        func = getattr(pwr, sys.argv[1])
        print(sys.argv[2:])
        func(*sys.argv[2:])
        # upd = pwr.pwr_get_outlets(upd_linked=True)
        # print(json.dumps(upd, indent=4, sort_keys=True))
else:
    pass
    # print(json.dumps(outlets, indent=4, sort_keys=True))
    # print(outlets['defined']['labpower1']['linked_devs'])

# for t in threading.enumerate():
#     print(t.name)
for e in log.error_msgs:
    print(e)



#!/etc/ConsolePi/venv/bin/python

import json
import readline # NoQA - allows input to accept backspace
import sys
import re
import threading
import itertools
from os import system
from pprint import pprint
from collections import OrderedDict as od
from typing import Union
from halo import Halo
import asyncio
# from rich.console import Console

# --// ConsolePi imports \\--
sys.path.insert(0, '/etc/ConsolePi/src/pypkg')
from consolepi.consolepi import ConsolePi  # type: ignore # NoQA
from consolepi.udevrename import Rename  # type: ignore # NoQA
from consolepi.menu import Menu  # type: ignore # NoQA
from consolepi import log, utils, config  # type: ignore # NoQA

# GoogleDrive import below is in refresh method of consolepi.remotes.Remotes(...).refresh
# package costs too much to import (time). Imported on demand when necessary
# from consolepi.gdrive import GoogleDrive

MIN_WIDTH = 55
MAX_COLS = 5


class Actions():
    def __init__(self, ):
        pass


# console = Console()


class Choice():
    def __init__(self, prompt: str, clear=False):
        if not clear:
            # console.begin_capture()
            # console.print(f"[bright_green]{prompt}[/]", end="")
            # prompt = console.end_capture()
            ch = input(prompt)
            self.lower = ch.lower()
            self.orig = ch
        else:
            self.lower = ''
            self.orig = ''


class ConsolePiMenu(Rename):

    def __init__(self, bypass_remotes: bool = False, bypass_outlets: bool = False):
        self.cpi = ConsolePi(bypass_remotes=bypass_remotes, bypass_outlets=bypass_outlets)
        self.cpiexec = self.cpi.cpiexec
        self.baud = config.default_baud
        self.go = True
        self.spin = Halo(spinner='dots')
        self.states = {
            True: '{{green}}ON{{norm}}',
            False: '{{red}}OFF{{norm}}'
        }
        self.log_sym_2bang = '\033[1;33m!!\033[0m'
        self.display_con_settings = False
        self.menu = self.cpi.menu
        self.menu.ignored_errors = [
            re.compile('Connection to .* closed'),
            re.compile('Warning: Permanently added .* to the list of known hosts')
        ]
        self.show_ports = False
        self.menu.page.hide = config.ovrd.get("hide_legend", False)
        self.do_menu_load_warnings()
        self.menu.legend_options = {
            'power': ['p', 'Power Control Menu'],
            'dli': ['d', '[dli] Web Power Switch Menu'],
            'rshell': ['rs', 'Remote Shell Menu'],
            'key': ['k', 'Distribute SSH public Key to Remote Hosts'],
            'shell': ['sh', 'Enter Local Shell'],
            'rn': ['rn', 'Rename Adapters'],
            'refresh': ['r', 'Refresh'],
            'sync': ['s', 'Sync with cloud'],
            'tp': ['tp', 'Toggle Display of associated TELNET ports'],
            'con': ['c', 'Change Default Serial Settings (devices marked with ** only)'],
            'picohelp': ['h', 'Display Picocom Help'],
            'back': ['b', 'Back'],
            'x': ['x', 'Exit']
        }
        self.cur_menu = None
        super().__init__(self.menu)

    def print_attribute(self, ch: str, locs: dict = {}) -> Union[bool, None]:
        '''Debugging Function allowing user to print class attributes / function returns.

        Params:
            ch {str}: the input from user prompt
            locs: locals() from calling function

        Returns:
            Bool | None -- True if ch is an attribute No return otherwise
        '''
        _attrs = [a for a in self.cpi.__dir__() if not a.startswith('_')]
        _attrs += ['cpi', 'this', 'self', 'menu', 'cloud', 'config', 'log']
        if '.' in ch and len(ch) > 2 and ch.split('.')[0] in _attrs:
            cpi = self.cpi
            cpiexec = cpi.cpiexec  # NoQA
            local = cpi.local  # NoQA
            if hasattr(cpi, 'remotes'):
                remotes = cpi.remotes
                cloud = remotes.cloud  # NoQA
            pwr = cpi.pwr  # NoQA
            menu = self.menu # NoQA
            _var = None

            if ' -pprint' in ch:
                do_pprint = True
                ch = ch.replace(' -pprint', '')
            else:
                do_pprint = False

            _key_str = _args_str = 'wtf?'
            _ch = ch
            if '[' in ch:
                _key_str = f"[{ch.split('[')[1]}"
                _ch = ch.replace(_key_str, '{{KEY}}')
            if '(' in ch:
                _args_str = f"({ch.split('(')[1]}"
                _ch = ch.replace(_args_str, '{{ARGS}}')
            _class_str = '.'.join(_ch.split('.')[0:-1])
            _attr = _ch.split('.')[-1].split('{{')[0]
            if _class_str.startswith('this'):
                _var = f'self.var = locs.get("{_attr}", "Attribute/Variable Not Found")'
                _var += ch.lstrip(f"{_class_str}.").lstrip(_attr)
            else:
                try:
                    exec(f"self._class = {_class_str}")
                    if hasattr(self._class, _attr):  # NoQA
                        _var = f"self.var = {ch}"
                    else:
                        log.show(f"DEBUG: '{_class_str}' object has no attribute '{_attr}'")
                        return True
                except AttributeError as e:
                    log.show(f'DEBUG: Attribute error {e}')

            if _var:
                try:
                    exec(_var)
                    if isinstance(self.var, (dict, list)):
                        _class_str = _class_str.replace('this', '')
                        if _class_str:
                            _class_str += '.'
                        if isinstance(self.var, dict):
                            var = {k: v if not callable(v) else f'{_class_str}{v.__name__}()' for k, v in self.var.items()}
                            for k in var:
                                try:
                                    if 'function' in var[k]:
                                        var[k]['function'] = f"{_class_str}{var[k]['function'].__name__}()"
                                except TypeError:
                                    continue
                        else:
                            var = [v if not callable(v) else f'{_class_str}{v.__name__}()' for v in self.var]

                        if not do_pprint:
                            print(json.dumps(var, indent=4, sort_keys=True))
                        else:
                            pprint(var)
                        print(f'{type(self.var)} length: {len(self.var)}')
                    else:
                        print(self.var)
                        print(f'type: {type(self.var)}\n'
                              if isinstance(self.var, str) and 'Not Found' not in self.var else '\n')
                    input('Press Enter to Continue... ')
                except Exception as e:
                    if hasattr(self, 'var'):
                        print(self.var)
                        input('Press Enter to Continue... ')
                    log.show(f'DEBUG: {e}')
                return True

    def do_menu_load_warnings(self):
        '''Displays warnings based on data collected/validated during menu load.'''

        # -- // Not running as root \\ --
        if not config.root:
            log.show('Running without sudo privs ~ Results may vary!\n'
                     'Use consolepi-menu to launch menu')

        # -- // No Local Adapters Found \\ --
        if not self.cpi.local.adapters:
            if not config.cloud_pull_only:
                log.show('No Local Adapters Detected')
            if log.error_msgs:
                # -- remove no ser2net.conf found msg if no local adapters
                log.error_msgs = [m for m in log.error_msgs if 'No ser2net configuration found' not in m]

    def picocom_help(self):
        print('----------------------- picocom Command Sequences -----------------------\n')
        print(' This program will launch serial session via picocom')
        print(' This is a list of the most common command sequences in picocom')
        print(' To use them press and hold ctrl then press and release each character\n')
        print('   ctrl+ a - x Exit session - reset the port')
        print('   ctrl+ a - q Exit session - without resetting the port')
        print('   ctrl+ a - b set baudrate')
        print('   ctrl+ a - u increase baud')
        print('   ctrl+ a - d decrease baud')
        print('   ctrl+ a - i change the number of databits')
        print('   ctrl+ a - f cycle through flow control options')
        print('   ctrl+ a - y cycle through parity options')
        print('   ctrl+ a - v Show configured port options')
        print('   ctrl+ a - c toggle local echo')
        print('\n-------------------------------------------------------------------------\n')
        input('Press Enter to Continue... ')

    # -- // POWER MENU \\ --
    def power_menu(self, calling_menu: str = 'main_menu'):
        cpi = self.cpi
        pwr = cpi.pwr
        # menu = self.menu
        menu = Menu("power_menu")
        states = self.states
        menu_actions = {
            # 'b': self.main_menu,
            'x': self.exit,
            'power_menu': self.power_menu,
        }
        menu.legend_options = {
            'power': ['p', 'Power Control Menu'],
            'dli': ['d', '[dli] Web Power Switch Menu'],
            'refresh': ['r', 'Refresh'],
            'back': ['b', 'Back'],
            'x': ['x', 'Exit']
        }
        show_linked = False
        choice = ''

        # Ensure Power Threads are complete
        if not cpi.pwr_init_complete:
            with Halo(text='Waiting for Outlet init threads to complete', spinner='dots'):
                self.cpiexec.wait_for_threads()
                outlets = pwr.data['defined']
        else:
            outlets = self.cpiexec.outlet_update()

        if not outlets:
            log.show('No Linked Outlets are connected')
            return
        if cpi.cpiexec.autopwr_wait:
            utils.spinner("Waiting for Auto Power Threads to Complete", cpi.cpiexec.wait_for_threads, name="auto_pwr", timeout=20)

        while choice not in ['x']:
            item = 1

            header = 'Power Control Menu'
            subhead = [
                '  enter item # to toggle power state on outlet',
                '  enter c + item # i.e. "c2" to cycle power on outlet'
            ]

            # Build menu items for each linked outlet
            state_list = []
            body = []
            _linked = ''
            legend = {'opts': [], 'before': []}
            for r in sorted(outlets):
                outlet = outlets[r]

                # -- // Linked DLI OUTLET AND ESPHOME MENU LINE(s) \\ --
                if outlet['type'].lower() in ['dli', 'esphome']:
                    is_on_dict = {}

                    # Show DLIs that are linked (Auto Power On)
                    if outlet['type'].lower() == "dli" and outlet.get('linked_devs') and outlet.get('is_on'):
                        is_on_dict = sorted(outlet['is_on'])

                    # Show linked esphome relays, or all if there is only 1.
                    elif len(outlet.get("relays", {})) == 1 or (outlet.get('linked_devs') and outlet.get('is_on')):
                        is_on_dict = {k: v for k, v in outlet["is_on"].items() if k in outlet.get("linked_devs", {}).values()}

                    for _port in is_on_dict:
                        if show_linked:
                            this_linked = config.cfg_yml.get('POWER', {}).get(r, {}).get('linked_devs', {})
                            _linked = [
                                '        {}'.format(
                                    '{{red}}' + k + '{{norm}}'
                                    if k not in cpi.local.adapters and '/host/' not in k else k
                                )
                                for k in this_linked if _port in utils.listify(this_linked[k])
                            ]
                            if _linked:
                                _linked.insert(0, '\n      {{cyan}}Linked Devices{{norm}}:')
                                _linked = '\n'.join(_linked).rstrip()

                        _outlet = outlet['is_on'][_port]
                        _state = states[_outlet['state']]

                        if not outlet.get("no_all", False):
                            state_list.append(_outlet['state'])

                        _name = ' ' + _outlet['name'] if 'ON' in _state else _outlet['name']

                        _menu_line = f"[{_state}] {_name}"
                        _menu_line = f"{_menu_line} ({r}" if _name.strip() != r else f"{_menu_line} ("
                        _menu_line = f"{_menu_line} Port:{_port})" if _name.strip() != _port else f"{_menu_line})"
                        _menu_line = f"{_menu_line} {_linked if _linked else ''}"
                        body.append(_menu_line)

                        menu_actions[str(item)] = {
                            'function': pwr.pwr_toggle,
                            'args': [outlet['type'], outlet['address']],
                            'kwargs': {'port': _port},  # , 'desired_state': not _outlet['state']},
                            'key': r
                        }
                        menu_actions['c' + str(item)] = {
                            'function': pwr.pwr_cycle,
                            'args': [outlet['type'].lower(), outlet['address']],
                            'kwargs': {'port': _port},
                            'key': 'dli_pwr' if outlet['type'].lower() == 'dli' else r
                        }
                        menu_actions['r' + str(item)] = {
                            'function': pwr.pwr_rename,
                            'args': [outlet['type'].lower(), outlet['address']],
                            'kwargs': {'port': _port},
                            'key': 'dli_pwr' if outlet['type'].lower() == 'dli' else r
                        }
                        item += 1

                # -- // GPIO or tasmota OUTLET MENU LINE \\ --
                else:
                    # pwr functions put any errors (aborts) in pwr.data[grp]['error']
                    if pwr.data['defined'][r].get('errors'):
                        log.show(f'{r} - {pwr.data["defined"][r]["errors"]}')
                        del pwr.data['defined'][r]['errors']

                    if show_linked:
                        this_linked = config.cfg_yml.get('POWER', {}).get(r, {}).get('linked_devs', {})
                        _linked = ['{}'.format('{{red}}' + k + '{{norm}}'
                                   if k not in cpi.local.adapters and '/host/' not in k else k)
                                   for k in this_linked]
                    if isinstance(outlet.get('is_on'), bool):
                        _state = states[outlet['is_on']]
                        state_list.append(outlet['is_on'])
                        body.append(f"[{_state}] {' ' + r if 'ON' in _state else r} ({outlet['type']}:{outlet['address']}) "
                                    f"{_linked if _linked else ''}")
                        menu_actions[str(item)] = {
                            'function': pwr.pwr_toggle,
                            'args': [outlet['type'], outlet['address']],
                            'kwargs': {'noff': outlet.get('noff', True)},
                            'key': r
                        }
                        menu_actions['c' + str(item)] = {
                            'function': pwr.pwr_cycle,
                            'args': [outlet['type'], outlet['address']],
                            'kwargs': {'noff': outlet.get('noff', True)},
                            'key': r
                        }
                        menu_actions['r' + str(item)] = {
                            'function': pwr.pwr_rename,
                            'args': [outlet['type'], outlet['address']],
                            'key': r
                        }
                        item += 1
                    else:   # refactored power.py pwr_get_outlets this should never hit
                        if outlet.get('is_on'):
                            log.show(f"DEV NOTE {r} outlet state is not bool: {outlet['error']}")

            if item > 2:
                if False in state_list:
                    legend['before'].append('all on: Turn all outlets {{green}}ON{{norm}}')
                    menu_actions['all on'] = {
                        'function': pwr.pwr_all,
                        'kwargs': {'outlets': outlets, 'action': 'toggle', 'desired_state': True}
                    }
                if True in state_list:
                    legend['before'].append('all off: Turn all outlets {{red}}OFF{{norm}}')
                    menu_actions['all off'] = {
                        'function': pwr.pwr_all,
                        'kwargs': {'outlets': outlets, 'action': 'toggle', 'desired_state': False}
                    }
                    legend['before'].append('cycle all: Cycle all outlets '
                                            '{{green}}ON{{norm}}{{dot}}{{red}}OFF{{norm}}{{dot}}{{green}}ON{{norm}}')

                    menu_actions['cycle all'] = {
                        'function': pwr.pwr_all,
                        'kwargs': {'outlets': outlets, 'action': 'cycle'}
                    }
                legend['before'].append('')
                if not show_linked:
                    legend['before'].append('L.  Show Connected Linked Devices')
                else:
                    legend['before'].append('L.  Hide Connected Linked Devices')

            legend['opts'] = ['back', 'refresh']
            if pwr.dli_exists and not calling_menu == 'dli_menu':
                legend['opts'].insert(0, 'dli')
                menu_actions['d'] = self.dli_menu

            menu_actions = menu.print_menu(body, header=header, subhead=subhead, legend=legend, menu_actions=menu_actions)
            choice_c = self.wait_for_input(locs=locals())
            choice = choice_c.lower
            if choice not in ['r', 'l']:
                if menu.cur_page == 1 and choice == "b":
                    break
                cpi.cpiexec.menu_exec(choice_c, menu_actions, calling_menu='power_menu')
            elif choice == 'l':
                show_linked = not show_linked
            elif choice == 'r':
                if pwr.dli_exists:
                    outlets = utils.spinner("Refreshing Outlets", self.cpiexec.outlet_update, refresh=True, upd_linked=True)

    def wait_for_input(self, prompt: str = " >> ", terminate: bool = False, locs: dict = {}) -> Choice:
        '''Get input from user.

        User Can Input One of the following for special handling:
        'debug': toggles debug (menu only doens't effect logging)
        'exit': exits the program
        a 3+ character string with '.' in it:  Which results in the print_attribute
        debug function being called.

        Other than exit, Any user input matching the above will result in a '' string
        being returned which results in a menu re-print.

        Keyword Arguments:
            prompt {str} -- input prompt (default: {" >> "})
            lower {bool} -- return lower case user input (default: {True})
            terminate {bool} -- terminates program if Ctrl+C/Ctrl+D (default: {False})
            locs {dict} -- locals from calling func for use in debug print func (default: {{}})

        Returns:
            str -- If user input is not 'exit' (which exits the program) will return an
                   empty str if pre-processed by one of the common funcs
                   otherwise returns the input provided by the user (lower() by default).
        '''
        menu = self.menu

        try:
            # if config.debug:
            #     prompt = f'pg[{menu.cur_page}] m[r:{menu.page.rows}, c:{menu.page.cols}] ' \
            #              f'a[r:{menu.tty.rows}, c:{menu.tty.cols}]{prompt}'
            if config.debug and self.cur_menu is not None:
                prompt = f'pg[{self.cur_menu.cur_page}] m[r:{self.cur_menu.page.rows}, c:{self.cur_menu.page.cols}] ' \
                         f'a[r:{self.cur_menu.tty.rows}, c:{self.cur_menu.tty.cols}]' \
                         f'{prompt}'

            ch = Choice(prompt)

            # -- // toggle debug \\ --
            if ch.lower == 'debug':
                log.show(f'debug toggled {self.states[config.debug]} --> {self.states[not config.debug]}')
                config.debug = not config.debug
                log.DEBUG = not log.DEBUG
                if config.debug:
                    log.setLevel(10)  # logging.DEBUG = 10
                ch = Choice(prompt, clear=True)

            # -- // always accept 'exit' \\ --
            elif ch.lower == 'exit':
                self.exit()

            # -- // Menu Debugging Tool prints attributes/function returns \\ --
            elif '.' in ch.orig and self.print_attribute(ch.orig, locs):
                ch = Choice(prompt, clear=True)

            menu.menu_rows = 0  # TODO REMOVE AFTER SIMPLIFIED

            return ch

        except (KeyboardInterrupt, EOFError):
            if terminate:
                print('Exiting based on User Input')
                self.exit()
            else:
                log.show('Operation Aborted')
                print('')  # prevents header and prompt on same line in debug
                return Choice(prompt, clear=True)

    # ------ // DLI WEB POWER SWITCH MENU / Multi-Port menu \\ ------ #
    def dli_menu(self, calling_menu: str = 'power_menu'):
        cpi = self.cpi
        pwr = cpi.pwr
        menu = Menu("dli_menu")
        menu.legend_options = {
            'power': ['p', 'Power Control Menu'],
            'refresh': ['r', 'Refresh'],
            'back': ['b', 'Back'],
            'x': ['x', 'Exit']
        }
        menu_actions = {
            'b': self.power_menu,
            'x': self.exit,
            'dli_menu': self.dli_menu,
            'power_menu': self.power_menu
        }

        choice = ''
        if not cpi.pwr_init_complete:
            with Halo(text='Waiting for Outlet init Threads to Complete...',
                      spinner='dots'):
                cpi.cpiexec.wait_for_threads('init')

        if not pwr.dli_exists:
            log.show('All Defined dli Web Power Switches are unreachable')
            return

        if cpi.cpiexec.autopwr_wait:
            utils.spinner("Waiting for Auto Power Threads to Complete", cpi.cpiexec.wait_for_threads, name="auto_pwr", timeout=20)

        while choice not in ['x']:
            item = start = 1
            outer_body = []
            slines = []
            state_list = []
            dli_dict = {**pwr.data['dli_power'], **pwr.data['esp_power']}
            for addr in sorted(dli_dict, key=lambda i: i.lower()):
                mlines = []
                state_list = []
                port_dict = dli_dict[addr]

                # strip off all but hostname if address is fqdn
                host_short = addr.split('.')[0] if '.' in addr and not addr.split('.')[0].isdigit() else addr

                # Menu includes both dli web power switches and multi-port esphome
                _type = "dli" if port_dict and str(list(port_dict.keys())[0]).isdigit() else "esphome"
                if _type == "dli":
                    key = "dli_pwr"
                    toggle_all_func = pwr.pwr_toggle
                    _all_args = [_type, addr]
                    toggle_all_kwargs = {'port': 'all'}
                    cycle_all_kwargs = {}
                    cycle_all_func = pwr.pwr_cycle
                else:
                    key = addr
                    toggle_all_func = pwr.pwr_all
                    defined = {k: v for k, v in pwr.data.get("defined", {}).items() if v.get("address", "") == addr}
                    if not defined or len(defined) > 1:
                        log.error(f"dli_menu unable to find def or too many matches for {addr}", show=True)
                    else:
                        key = list(defined.keys())[0]
                    _all_args = [defined]
                    toggle_all_kwargs = {"action": "toggle"}
                    cycle_all_kwargs = {"action": "cycle"}
                    cycle_all_func = pwr.pwr_all

                # -- // MENU ITEMS LOOP \\ --
                for idx, port in enumerate(port_dict):
                    pname = port_dict[port]['name']
                    cur_state = port_dict[port]['state']
                    state_list.append(cur_state)
                    to_state = not cur_state
                    on_pad = ' ' if cur_state else ''
                    # _type = "dli" if port.isdigit() else "esphome"

                    # TODO make esp-power it's own dict from pwr_get_outlets vs putting in dli-power
                    if _type == "dli":
                        mlines.append('[{}] {}{}{}'.format(
                            self.states[cur_state], on_pad, 'P' + str(port) + ': ' if port != item else '', pname))

                    else:  # esphome in dli menu
                        mlines.append('[{}] {}{}{}'.format(self.states[cur_state], on_pad, 'P' + str(idx + 1) + ': ' if idx != item else '', pname))
                        _type = "esphome"

                    menu_actions[str(item)] = {
                        'function': pwr.pwr_toggle,
                        'args': [_type, addr],
                        'kwargs': {'port': port},  # , 'desired_state': to_state},
                        'key': key
                    }
                    menu_actions['c' + str(item)] = {
                        'function': pwr.pwr_cycle,
                        'args': [_type, addr],
                        'kwargs': {'port': port},
                        'key': key
                    }
                    menu_actions['r' + str(item)] = {
                        'function': pwr.pwr_rename,
                        'args': [_type, addr],
                        'kwargs': {'port': port},
                        'key': key
                    }

                    item += 1

                # add final entry for all operations
                if True not in state_list:
                    _line = 'ALL {{green}}ON{{norm}}'
                    desired_state = True
                elif False not in state_list:
                    _line = 'ALL {{red}}OFF{{norm}}'
                    desired_state = False
                else:
                    _line = 'ALL [on|off]. i.e. "{} off"'.format(item)
                    desired_state = None

                # build appropriate menu_actions item will represent ALL ON or ALL OFF if current state
                # of all outlets is the inverse
                # if there is a mix item#+on or item#+off will both be valid but item# alone will not.
                if desired_state in [True, False]:
                    menu_actions[str(item)] = {
                        'function': toggle_all_func,
                        'args': _all_args,
                        'kwargs': {**toggle_all_kwargs, 'desired_state': desired_state},
                        'key': key
                    }
                elif desired_state is None:
                    for s in ['on', 'off']:
                        desired_state = True if s == 'on' else False
                        menu_actions[str(item) + ' ' + s] = {
                            'function': toggle_all_func,
                            'args': _all_args,
                            'kwargs': {**toggle_all_kwargs, 'desired_state': desired_state},
                            'key': key
                        }
                mlines.append(_line)
                item += 1

                # Add cycle line if any outlets are currently ON
                if True in state_list:
                    mlines.append('Cycle ALL')
                    menu_actions[str(item)] = {
                        'function': cycle_all_func,
                        'args': _all_args,
                        'kwargs': cycle_all_kwargs,
                        'key': key
                    }
                item = start + 10
                start += 10

                outer_body.append(mlines)   # list of lists where each list = printed menu lines
                slines.append(host_short)   # list of strings index to index match with body list of lists

            header = 'DLI Web Power Switch / espHome Power Strip'
            subhead = ['enter item # to toggle power state on outlet']
            if True in state_list:
                subhead.append('enter c + item # i.e. "c2" to cycle power on outlet')
            subhead.append('enter r + item # i.e. "r2" to rename the outlet')

            legend = {'opts': ['back', 'refresh']}
            if (not calling_menu == 'power_menu' and pwr.data) and (pwr.gpio_exists or pwr.tasmota_exists or pwr.linked_exists or pwr.esphome_exists):
                menu_actions['p'] = self.power_menu
                legend['opts'].insert(0, 'power')
                legend['overrides'] = {'power': ['p', 'Power Control Menu (linked, GPIO, tasmota)']}

            # for dli menu remove any tasmota errors
            for _error in log.error_msgs:
                if 'TASMOTA' in _error:
                    log.error_msgs.remove(_error)

            menu_actions = menu.print_menu(outer_body, header=header, subhead=subhead, legend=legend, subs=slines,
                                           by_tens=True, menu_actions=menu_actions)

            self.cur_menu = menu

            if menu.cur_page == 1 and choice == "b":
                break

            choice_c = self.wait_for_input(locs=locals())
            choice = choice_c.lower
            if choice == 'r':
                self.spin.start('Refreshing Outlets')
                dli_dict = self.cpiexec.outlet_update(refresh=True, key='dli_power')
                esp_dict = self.cpiexec.outlet_update(refresh=True, key='esp_power')
                pwr.data['dli_power'] = dli_dict
                pwr.data['esp_power'] = esp_dict
                self.spin.succeed()
            elif choice == 'b':
                return
            else:
                cpi.cpiexec.menu_exec(choice_c, menu_actions, calling_menu='dli_menu')

    def key_menu(self):
        cpi = self.cpi
        rem = cpi.remotes.data
        menu = Menu()
        menu.legend_options = {
            'refresh': ['r', 'Refresh'],
            'back': ['b', 'Back'],
            'x': ['x', 'Exit']
        }
        choice = ''
        menu_actions = {
            'b': None,
            'x': self.exit,
            'key_menu': self.key_menu
        }
        while choice.lower() not in ['x']:
            header = ' Remote SSH Key Distribution Menu '
            subhead = ["Use this menu to distribute your ssh public key to remote hosts.",
                       "SSH Public/Private keys are generated if one doesn't already exist.",
                       "This facilitates certificate based authentication."]

            # Build menu items for each serial adapter found on remote ConsolePis
            item = 1
            outer_body = []
            all_list = []
            mlines = []
            subs = ['Remote ConsolePis']
            for host in sorted(rem):
                if 'rem_ip' not in rem[host]:
                    log.warning('[KEY_MENU] {} lacks rem_ip skipping'.format(host))
                    continue

                rem_ip = rem[host].get('rem_ip')
                if rem_ip:
                    rem_user = rem[host].get('user', config.static.get('FALLBACK_USER', 'pi'))
                    mlines.append(f'{host} @ {rem_ip}')
                    this = (host, rem_ip, rem_user)
                    menu_actions[str(item)] = {'function': self.cpiexec.gen_copy_key, 'args': [this]}
                    all_list.append(this)
                    item += 1

            outer_body.append(mlines)

            # Build menu items for each manually defined host in ConsolePi.yaml
            host_all_list = []
            if config.hosts:
                for _key in config.hosts:
                    if not _key.startswith('_'):
                        for _sub in config.hosts[_key]:
                            host_mlines = []
                            for host in config.hosts[_key][_sub]:
                                this = config.hosts[_key][_sub][host]
                                if "address" in this and "username" in this and this.get("method", "").lower() == "ssh":
                                    host_mlines.append(f"{host.replace('/host/', '')} @ {this['address']}")
                                    this_args = (host.replace('/host/', ''), this['address'], this['username'])
                                    menu_actions[str(item)] = {'function': self.cpiexec.gen_copy_key, 'args': [this_args]}
                                    host_all_list.append(this_args)
                                    item += 1
                            if host_mlines:
                                subs.append(_sub)
                                outer_body.append(host_mlines)

            # -- all option to loop through all remotes and deploy keys --
            legend = {'opts': 'back',
                      'before': ['all cpis:  {{cyan}}*all*{{norm}} Remote ConsolePis',
                                 'all hosts: {{cyan}}*all*{{norm}} Manual Host Entries',
                                 '           {{dyellow}}**host must support ssh-copy-id{{norm}}',
                                 ''
                                 ]}

            menu_actions['all cpis'] = {'function': self.cpiexec.gen_copy_key, 'args': [all_list]}
            menu_actions['all hosts'] = {'function': self.cpiexec.gen_copy_key, 'args': [host_all_list]}

            # menu_actions = self.menu.print_menu(outer_body, subs=subs, header=header, subhead=subhead,
            #                                     legend=legend, menu_actions=menu_actions)
            menu_actions = menu.print_menu(outer_body, subs=subs, header=header, subhead=subhead,
                                           legend=legend, menu_actions=menu_actions)
            self.cur_menu = menu
            choice_c = self.wait_for_input(locs=locals())
            choice = choice_c.lower

            # TODO Temp need more elegant way to handle back to main_menu
            if menu.cur_page == 1 and choice == "b":
                break

            cpi.cpiexec.menu_exec(choice_c, menu_actions, calling_menu='key_menu')

    def gen_adapter_lines(self, adapters: dict, item: int = 1, remote: bool = False,
                          rem_user: str = None, host: str = None, rename: bool = False) -> tuple:
        cpi = self.cpi
        if hasattr(cpi, 'remotes'):
            rem = cpi.remotes.data
        else:
            rem = {}
        menu_actions = {}
        mlines = []
        flow_pretty = self.flow_pretty

        # -- // Manually Defined Hosts \\ --
        if adapters.get('_hosts'):
            for h in adapters['_hosts']:
                menu_line = adapters['_hosts'][h].get('menu_line')
                _m = "ssh"  # init.  TODO verify, and remove, shouldn't need to build the cmd below, built in config
                host_pretty = h  # init
                if not menu_line:
                    host_pretty = h.replace('/host/', '')
                    _addr = adapters['_hosts'][h]['address']
                    _m = adapters['_hosts'][h].get('method')
                    if ':' in _addr:
                        _a, _p = _addr.split(':')
                        _a = utils.get_host_short(_a)
                        _addr = _a if (_m == 'telnet' and _p == '23') or \
                            (_m == 'ssh' and _p == '22') else f'{_a}:{_p}'
                    menu_line = f'{host_pretty} @ {_addr}'
                mlines.append(menu_line)
                menu_actions[str(item)] = {'cmd': adapters['_hosts'][h].get('cmd'),
                                           'exec_kwargs': {'tee_stderr': True}, 'pwr_key': h,
                                           'pre_msg': f"Establishing {_m} session To {host_pretty}..."}
                item += 1

            return mlines, menu_actions, item

        # -- // Local Adapters \\ --
        for _dev in sorted(adapters.items(), key=lambda i: i[1]['config'].get('port', 0)):
            dev = _dev[0]
            cfg_dict = adapters[dev].get('config', adapters[dev])
            port = cfg_dict.get('port')
            if port != 0:
                def_indicator = ''
            else:
                def_indicator = '**'
                self.display_con_settings = True  # Can Remove - no longer used, use picocom if no alias
            baud = cfg_dict.get('baud', self.baud)
            dbits = cfg_dict.get('dbits', 8)
            flow = cfg_dict.get('flow', 'n')
            parity = cfg_dict.get('parity', 'n')
            sbits = cfg_dict.get('sbits', 1)
            dev_pretty = dev.replace('/dev/', '')

            # Generate Adapter Menu Line
            if not self.show_ports:
                menu_line_sfx = f'{def_indicator}[{baud} {dbits}{parity[0].upper()}{sbits}]'
            else:
                menu_line_sfx = '**undefined' if port == 0 else port

            if flow != 'n' and flow in flow_pretty:
                menu_line_sfx += f' {flow_pretty[flow]}'
            menu_line = f'{dev_pretty} {menu_line_sfx}'
            mlines.append(menu_line)

            # fallback_cmd should never be used as the cmd should always be in the dev dict
            loc_cmd = f'picocom {dev} --baud {baud} --flow {flow} --databits {dbits} --parity {parity}'
            # -- // Adapter menu_actions \\ --
            if not remote:
                # Generate connect command used to connect to device
                _cmd = cfg_dict.get('cmd', loc_cmd)
                if not rename:
                    menu_actions[str(item)] = {'cmd': _cmd, 'pwr_key': dev,
                                               'pre_msg': f"Connecting To {dev_pretty}..."}
                else:
                    rn_this = {dev: adapters[dev]}
                    menu_actions[str(item)] = {'function': self.do_rename_adapter, 'args': [dev]}
                    menu_actions['s' + str(item)] = {'function': self.cpiexec.show_adapter_details, 'args': [rn_this]}
                    menu_actions['c' + str(item)] = {'cmd': _cmd, 'pre_msg': f"Connecting To {dev_pretty}..."}

            # -- // REMOTE ADAPTERS \\ --
            else:
                if not rem_user:  # the user advertised by the remote we ssh to the remote with this user
                    rem_user = config.FALLBACK_USER

                # Generate connect command used to connect to remote device
                # TODO simplify once api updated to use new libraries
                rem_pfx = f'sudo -u {config.loc_user} ssh -t {rem_user}@{rem[host].get("rem_ip")}'
                fallback_cmd = f'{rem_pfx} \"{config.REM_LAUNCH} {loc_cmd}\"'
                _cmd = cfg_dict.get('cmd', fallback_cmd)
                if 'ssh -t' not in _cmd:
                    _cmd = f'{rem_pfx} \"{config.REM_LAUNCH} {_cmd}\"'

                connect = {'cmd': _cmd, 'pre_msg': f"Connecting To {dev_pretty} on {host}..."}
                if not rename:
                    menu_actions[str(item)] = connect
                else:
                    menu_actions['c' + str(item)] = connect
                    menu_actions['c ' + str(item)] = connect

                    _cmd = f'{rem_pfx} \"sudo /\etc/\ConsolePi/\src/\consolepi-menu.py rn {dev}\"'  # type: ignore # noqa
                    menu_actions[str(item)] = {'cmd': _cmd,
                                               'pre_msg': f"Connecting To {host} to Rename {dev_pretty}...",
                                               'host': host}
                    rn_this = {dev: adapters[dev]}
                    if rn_this[dev].get('udev'):
                        menu_actions['s' + str(item)] = {'function': self.cpiexec.show_adapter_details, 'args': [rn_this]}
                        menu_actions['s ' + str(item)] = {'function': self.cpiexec.show_adapter_details, 'args': [rn_this]}
                    else:
                        _msg = 'Adapter Attributes not found in remote data, this feature is still in dev for remotes.'
                        menu_actions['s' + str(item)] = {'function': log.show, 'args': [_msg]}
                        menu_actions['s ' + str(item)] = {'function': log.show, 'args': [_msg]}

            item += 1

        return mlines, menu_actions, item

    def rename_menu(self, direct_launch: bool = False, from_name: str = None):
        cpi = self.cpi
        menu = Menu(name='rename_menu')
        menu.legend_options = {
            'refresh': ['r', 'Refresh'],
            'tp': ['tp', 'Toggle Display of associated TELNET ports'],
            'back': ['b', 'Back'],
            'x': ['x', 'Exit']
        }
        local = cpi.local
        remotes = cpi.remotes if not direct_launch else None
        choice = ''
        menu_actions = {}
        # -- rename invoked from another ConsolePi (remote rename) --
        if direct_launch and from_name:
            self.do_rename_adapter(from_name)
            self.trigger_udev()
            sys.exit()
        while choice not in ["x"]:
            if choice == 'r':
                local.adapters = local.build_adapter_dict(refresh=True)
                if not direct_launch:
                    remotes.data = asyncio.run(remotes.get_remote(data=config.remote_update()))
            loc = local.adapters
            rem = remotes.data if not direct_launch else []

            if (direct_launch and not loc) or (not loc and not rem):
                print(f"{self.log_sym_2bang}  No Local Adapters Detected. Nothing to rename, exiting...")
                return

            slines = []
            outer_body = []
            item = 1

            if loc:
                slines.append('Rename Local Adapters')   # list of strings index to index match with body list of lists
                mlines, menu_actions, item = self.gen_adapter_lines(loc, rename=True)
                outer_body.append(mlines)
            rem_item = item

            if not direct_launch:
                for host in remotes.data:
                    if rem[host].get('rem_ip') and rem[host].get('adapters'):
                        slines.append(f'Rename Adapters on {host}')
                        mlines, rem_menu_actions, item = self.gen_adapter_lines(rem[host]['adapters'], item=item,
                                                                                remote=True, rem_user=rem[host].get('user'),
                                                                                host=host, rename=True)
                        outer_body.append(mlines)
                        menu_actions = {**menu_actions, **rem_menu_actions}

            legend = {'before': [
                's#. Show details for the adapter i.e. \'s1\'',
                'c#. Connect to the device i.e. \'c1\'',
                '',
                'The system is updated with new adapter(s) when you leave this menu.',
                ''
            ], 'opts': []}
            if not direct_launch:
                legend['opts'].append('back')

            legend['opts'].append('tp')
            menu_actions['tp'] = self.toggle_show_ports
            legend['opts'].append('refresh')
            menu_actions['b'] = None
            menu_actions['r'] = None
            menu_actions['x'] = self.exit

            menu_actions = menu.print_menu(outer_body, header='Define/Rename Adapters',
                                           legend=legend, subs=slines, menu_actions=menu_actions)

            self.cur_menu = menu

            choice_c = self.wait_for_input(locs=locals())
            choice = choice_c.lower

            # if trying to connect to local adapter after rename refresh udev
            if choice.startswith('c') and len(choice) <= len(str(rem_item - 1)) + 1 and self.udev_pending:
                n = int(choice.replace('c', '').strip())
                if n < rem_item:
                    self.trigger_udev()

            cpi.cpiexec.menu_exec(choice_c, menu_actions, calling_menu='rename_menu')

            # -- if rename was performed on a remote update remotes to pull the new name
            if choice.isdigit() and int(choice) >= rem_item:
                print('Triggering Refresh due to Remote Name Change')
                # remotes.refresh(bypass_cloud=True)  # NoQA TODO would be more ideal just to query the remote involved in the rename and update the dict
                remotes.data = asyncio.run(remotes.get_remote(data=config.remote_update(), rename=True))

            # TODO Temp need more elegant way to handle back to main_menu
            elif menu_actions.get(choice, {}) is None and choice == "b":
                break

        # trigger refresh udev and restart ser2net after rename
        if self.udev_pending:
            self.trigger_udev()

    def toggle_show_ports(self):
        self.show_ports = not self.show_ports

    def refresh_local(self):
        cpi = self.cpi
        remotes = cpi.remotes
        cpi.local.adapters = cpi.local.build_adapter_dict(refresh=True)
        remotes.data = asyncio.run(remotes.get_remote(data=config.remote_update()))

    # ------ // MAIN MENU \\ ------ #
    def main_menu(self):
        cpi = self.cpi
        menu = cpi.menu
        menu.name = "main_menu"
        loc = cpi.local.adapters
        pwr = cpi.pwr
        remotes = cpi.remotes
        rem = cpi.remotes.data
        outer_body = []
        slines = []
        item = 1
        foot_opts = []

        menu_actions = {
            'h': self.picocom_help,
            'r': remotes.refresh,
            'rl': self.refresh_local,
            'x': self.exit,
        }
        if config.power and pwr.data:
            if pwr.linked_exists or pwr.gpio_exists or pwr.tasmota_exists:
                menu_actions['p'] = self.power_menu
            elif pwr.dli_exists:  # if no linked outlets but dlis "p" also sends to dli_menu (valid but not displayed)
                menu_actions['p'] = self.dli_menu

            if pwr.dli_exists:
                menu_actions['d'] = self.dli_menu

        # Direct launch to power menu's if nothing to show in main and power enabled.
        if not loc and not rem and (not config.hosts or not config.hosts.get('main')) and config.power:
            log.show('No Adapters Found, Outlets Defined... Launching to Power Menu\n'
                     'use option "b" to access main menu options')
            if pwr.dli_exists and not pwr.linked_exists:
                cpi.cpiexec.menu_exec('d', menu_actions)
            else:
                cpi.cpiexec.menu_exec('p', menu_actions)

        # Build menu items for each locally connected serial adapter
        if loc:
            mlines, loc_menu_actions, item = self.gen_adapter_lines(loc)
            if loc_menu_actions:
                menu_actions = {**loc_menu_actions, **menu_actions}

            outer_body.append(mlines)   # list of lists where each list = printed menu lines
            slines.append('[LOCAL] Directly Connected')   # Sub-heads: list of str idx to idx match with outer_body list of lists

        # Build menu items for each serial adapter found on remote ConsolePis
        rem_mlines = []
        rem_slines = []
        rem_outer_body = []
        for host in sorted(rem):
            if rem[host].get('rem_ip') and rem[host]['adapters']:
                remotes.connected = True
                rem_mlines, rem_menu_actions, item = self.gen_adapter_lines(rem[host]['adapters'], item=item, remote=True,
                                                                            rem_user=rem[host].get('user'), host=host)
                if rem_menu_actions:
                    menu_actions = {**menu_actions, **rem_menu_actions}

                rem_outer_body.append(rem_mlines)
                rem_slines.append('[Remote] {} @ {}'.format(host, rem[host]['rem_ip']))

        # -- // COMPACT MODE \\ --
        if remotes.connected:
            if config.compact_mode:
                slines.append('[Remote] On Remote ConsolePis')
                outer_body.append(list(itertools.chain.from_iterable(rem_outer_body)))
            else:
                slines += rem_slines
                outer_body += rem_outer_body

        # Build menu items for each manually defined host with show_in_main = True
        # TODO add background reachability check
        if config.hosts and config.hosts.get('main'):
            hosts = config.hosts['main']
            for g in hosts:
                mlines, host_menu_actions, item = self.gen_adapter_lines({'_hosts': hosts[g]}, item=item, host=g)
                if host_menu_actions:
                    menu_actions = {**menu_actions, **host_menu_actions}

                outer_body.append(mlines)
                slines.append(g)

        if loc or remotes.connected:
            foot_opts.append('picohelp')
            foot_opts.append('tp')
            menu_actions['tp'] = self.toggle_show_ports

        if config.power:  # and config.outlets is not None:
            if pwr.outlets_exists:
                foot_opts.append('power')
            if pwr.dli_exists:
                foot_opts.append('dli')

        if remotes.connected or config.hosts:
            if remotes.connected:
                foot_opts.append('key')
                menu_actions['k'] = self.key_menu

            foot_opts.append('rshell')
            menu_actions['rs'] = self.rshell_menu

        # foot_opts.append('shell') # Not really needed can just do bash -l from menu
        if loc or remotes.connected:  # and config.root:
            foot_opts.append('rn')
            menu_actions['rn'] = self.rename_menu
        foot_opts.append('refresh')

        menu_actions = menu.print_menu(
            outer_body, header='{{cyan}}Console{{red}}Pi{{norm}} {{cyan}}Serial Menu{{norm}}',
            legend={'opts': foot_opts}, subs=slines, menu_actions=menu_actions
        )

        self.cur_menu = menu

        choice_c = self.wait_for_input(locs=locals(), terminate=True)

        cpi.cpiexec.menu_exec(choice_c, menu_actions)

        return

    # ------ // REMOTE SHELL MENU \\ ------ #
    def rshell_menu(self):
        choice = ''
        cpi = self.cpi
        local = cpi.local
        rem = cpi.remotes.data
        menu = Menu("rshell_menu")
        menu.legend_options = {
            'refresh': ['r', 'Refresh'],
            'back': ['b', 'Back'],
            'x': ['x', 'Exit']
        }
        menu_actions = {
            'rshell_menu': self.rshell_menu,
            'b': None,
            'x': self.exit
        }

        while choice not in ['x']:
            outer_body = []
            mlines = []
            subs = []
            item = 1

            # Build menu items for each reachable remote ConsolePi
            subs.append('Remote ConsolePis')
            # TODO make a sep method as main essentially has this same section
            for host in sorted(rem):
                if rem[host].get('rem_ip'):
                    mlines.append(f'{host} @ {rem[host]["rem_ip"]}')
                    _rem_user = rem[host].get('user', config.FALLBACK_USER)
                    _cmd = f'sudo -u {local.user} ssh -t {_rem_user}@{rem[host]["rem_ip"]}'
                    menu_actions[str(item)] = {'cmd': _cmd,
                                               'pre_msg': f'Establishing ssh session To {host}...'}
                    item += 1
            outer_body.append(mlines)

            # Build menu items for each manually defined host in ConsolePi.yaml
            if config.hosts:
                for _sub in config.hosts['rshell']:
                    subs.append(_sub)
                    ssh_hosts = config.hosts['rshell'][_sub]
                    mlines, host_menu_actions, item = self.gen_adapter_lines({'_hosts': ssh_hosts}, item=item)
                    menu_actions = {**menu_actions, **host_menu_actions}
                    outer_body.append(mlines)

            legend = {'opts': 'back'}
            menu_actions = menu.print_menu(outer_body, header='Remote Shell Menu',
                                           subhead='Enter item # to connect to remote',
                                           legend=legend, subs=subs, menu_actions=menu_actions)

            choice_c = self.wait_for_input(locs=locals())
            choice = choice_c.lower

            # if choice == "b":
            #     break

            cpi.cpiexec.menu_exec(choice_c, menu_actions, calling_menu='rshell_menu')

            # TODO Temp need more elegant way to handle back to main_menu
            if menu_actions.get(choice, {}) is None and choice == "b":
                break

    # -- // CONNECTION MENU \\ --
    def con_menu(self, rename: bool = False, con_dict: dict = None):
        # menu = self.cpi.menu
        menu = Menu("con_menu")
        # menu.legend_options = {
        #     'back': ['b', 'Back'],
        #     'x': ['x', 'Exit']
        # }
        menu_actions = {
            '1': self.baud_menu,
            '2': self.data_bits_menu,
            '3': self.parity_menu,
            '4': self.flow_menu,
            'b': self.main_menu if not rename else self.do_rename_adapter,  # not used currently
            'x': self.exit
        }
        if con_dict:
            self.con_dict = {
                'baud': self.baud,
                'data_bits': self.data_bits,
                'parity': self.parity,
                'flow': self.flow,
                'sbits': self.sbits
            }
            self.baud = con_dict['baud']
            self.data_bits = con_dict['data_bits']
            self.parity = con_dict['parity']
            self.flow = con_dict['flow']
            self.sbits = con_dict['sbits']
        while True:
            header = 'Connection Settings Menu'
            mlines = []
            mlines.append('Baud [{}]'.format(self.baud))
            mlines.append('Data Bits [{}]'.format(self.data_bits))
            mlines.append('Parity [{}]'.format(self.parity_pretty[self.parity]))
            mlines.append('Flow [{}]'.format(self.flow_pretty[self.flow]))
            legend = {'opts': ['back', 'x'],
                      'overrides': {'back': ['b', 'Back {}'.format(' (Apply Changes to Files)' if rename else '')]}
                      }
            menu.print_menu(mlines, header=header, legend=legend, menu_actions=menu_actions, hide_legend=False)
            ch = self.wait_for_input(locs=locals()).lower
            try:
                if ch == 'b':
                    break
                menu_actions[ch]()  # could parse and store if decide to use this for something other than rename

            except KeyError as e:
                if ch:
                    log.show(f'Invalid selection {e}, please try again.')
                    log.clear()

    # -- // BAUD MENU \\ --
    def baud_menu(self):
        # config = self.cpi.config
        # menu = self.cpi.menu
        menu = Menu("baud_menu")
        menu_actions = od([
            ('1', 300),
            ('2', 1200),
            ('3', 9600),
            ('4', 19200),
            ('5', 57600),
            ('6', 115200),
            ('c', 'custom')
        ])
        # text = ' b.  Back'
        std_baud = [110, 300, 600, 1200, 2400, 4800, 9600, 14400, 19200, 38400, 57600, 115200, 128000, 256000]

        while True:
            # -- Print Baud Menu --
            if not config.debug:
                _ = system("clear -x")
            header = menu.format_header(text=' Select Desired Baud Rate ')
            print(header)

            for key in menu_actions:
                _cur_baud = menu_actions[key]
                print(' {0}. {1}'.format(key, _cur_baud if _cur_baud != self.baud else '[{}]'.format(_cur_baud)))

            legend = menu.format_legend(legend={"opts": 'back'})
            menu.page.legend.hide = False  # TODO make elegant this prevents "Use 'TL' to Toggle Legend from being displayed in footer given we are not giving an option in this menu"
            footer = menu.format_footer()
            print(legend)
            print(footer)
            log.clear()
            ch = self.wait_for_input(" Baud >>  ", locs=locals()).lower

            # -- Evaluate Response --
            try:
                if ch == 'c':
                    while True:
                        self.baud = self.wait_for_input(' Enter Desired Baud Rate >> ').lower
                        if not self.baud.isdigit():
                            print('Invalid Entry {}'.format(self.baud))
                        elif int(self.baud) not in std_baud:
                            _choice = input(' {} is not a standard baud rate are you sure? (y/n) >> '.format(self.baud)).lower()
                            if _choice not in ['y', 'yes']:
                                break
                        else:
                            break
                elif ch == 'b':
                    break  # return to con_menu
                elif ch == 'x':
                    self.exit()
                else:
                    self.baud = menu_actions[ch]
                    break
            except KeyError as e:
                if ch:
                    log.show('Invalid selection {} please try again.'.format(e))

        return self.baud

    # -- // DATA BITS MENU \\ --
    def data_bits_menu(self):
        # menu = self.cpi.menu
        menu = Menu("data_bits")
        valid = False
        while not valid:
            if not config.debug:
                _ = system("clear -x")
            header = menu.format_header(text=' Enter Desired Data Bits ')
            print(header)
            print('\n Default 8, Current [{}], Valid range 5-8'.format(self.data_bits))

            legend = menu.format_legend(legend={"opts": 'back'})
            menu.page.legend.hide = False  # TODO make elegant this prevents "Use 'TL' to Toggle Legend from being displayed in footer given we are not giving an option in this menu"
            footer = menu.format_footer()
            print(legend)
            print(footer)
            log.clear()
            choice = self.wait_for_input(' Data Bits >>  ', locs=locals()).orig
            try:
                if choice.lower() == 'x':
                    self.exit()
                elif choice.lower() == 'b':
                    valid = True
                elif int(choice) >= 5 and int(choice) <= 8:
                    self.data_bits = choice
                    valid = True
                else:
                    log.show('Invalid selection {} please try again.'.format(choice))
            except ValueError:
                if choice:
                    log.show('Invalid selection {} please try again.'.format(choice))

        return self.data_bits

    # -- // PARITY MENU \\ --
    def parity_menu(self):
        menu = Menu("parity_menu")

        def print_menu():
            if not config.debug:
                _ = system("clear -x")
            header = menu.format_header(text=' Select Desired Parity ')
            print(header)
            print('\n Default No Parity\n')
            print(f" 1. {'[None]' if self.parity == 'n' else 'None'}")
            print(f" 2. {'[Odd]' if self.parity == 'o' else 'Odd'}")
            print(f" 3. {'[Even]' if self.parity == 'e' else 'Even'}")

            legend = menu.format_legend(legend={"opts": 'back'})
            menu.page.legend.hide = False  # TODO make elegant this prevents "Use 'TL' to Toggle Legend from being displayed in footer given we are not giving an option in this menu"
            footer = menu.format_footer()
            print(legend)
            print(footer)
            log.clear()
        valid = False
        while not valid:
            print_menu()
            valid = True
            choice = self.wait_for_input(' Parity >> ', locs=locals()).lower
            if choice == '1':
                self.parity = 'n'
            elif choice == '2':
                self.parity = 'o'
            elif choice == '3':
                self.parity = 'e'
            elif choice == 'b':
                pass
            elif choice == 'x':
                self.exit()
            else:
                valid = False
                if choice:
                    log.show('Invalid selection {} please try again.'.format(choice))
        return self.parity

    # -- // FLOW MENU \\ --
    def flow_menu(self):
        menu = Menu("flow_menu")

        def print_menu():
            if not config.debug:
                _ = system("clear -x")
            header = menu.format_header(text=' Select Desired Flow Control ')
            print(header)
            print('')
            print(' Default No Flow\n')
            print(f" 1. {'[None]' if self.flow == 'n' else 'None'}")
            print(f" 2. {'[Xon/Xoff]' if self.flow == 'x' else 'Xon/Xoff'} (software)")
            print(f" 3. {'[RTS/CTS]' if self.flow == 'h' else 'RTS/CTS'} (hardware)")
            legend = menu.format_legend(legend={"opts": 'back'})
            menu.page.legend.hide = False  # TODO make elegant this prevents "Use 'TL' to Toggle Legend from being displayed in footer given we are not giving an option in this menu"
            footer = menu.format_footer()
            print(legend)
            print(footer)
            log.clear()
        valid = False
        while not valid:
            print_menu()
            choice = self.wait_for_input(' Flow >>  ', locs=locals()).lower
            if choice in ['1', '2', '3', 'b', 'x']:
                valid = True
            try:
                if choice == '1':
                    self.flow = 'n'
                elif choice == '2':
                    self.flow = 'x'
                elif choice == '3':
                    self.flow = 'h'
                elif choice == 'b':
                    pass
                elif choice == 'x':
                    self.exit()
                else:
                    if choice:
                        log.show('Invalid selection {} please try again.'.format(choice))
            except Exception as e:
                if choice:
                    log.show('Invalid selection {} please try again.'.format(e))

        return self.flow

    # -- // EXIT \\ --
    def exit(self):
        self.go = False

        cpi = self.cpi
        if config.power:
            if not cpi.pwr_init_complete:
                utils.spinner('Exiting... Waiting for Outlet init threads to complete',
                              self.cpiexec.wait_for_threads)
            if cpi.pwr and cpi.pwr._dli:
                threading.Thread(target=cpi.pwr.dli_close_all).start()

        # - if exit directly from rename menu after performing a rename trigger / reload udev
        if self.udev_pending:
            error = self.trigger_udev()
            if error:
                print(error)

        sys.exit(0)


if __name__ == '__main__':
    if len(sys.argv) > 1:
        # -- // Direct Launch to Rename Menu \\ --
        if sys.argv[1].lower() in ['rn', 'rename', 'addconsole']:
            cpi_menu = ConsolePiMenu(bypass_remotes=True, bypass_outlets=True)
            if len(sys.argv) == 2:
                cpi_menu.rename_menu(direct_launch=True)
            # -- // Direct Launch to Rename Task (Remote Rename) \\
            else:
                cpi_menu.do_rename_adapter(from_name=sys.argv[2])
                if cpi_menu.udev_pending:
                    error = cpi_menu.trigger_udev()
                    if error:
                        print(error, file=sys.stderr)
        # -- // Attribute Printer (used for debug and to see data structures) \\ --
        else:
            cpi_menu = ConsolePiMenu(bypass_remotes=True)
            var_in = sys.argv[1].replace('self', 'menu')
            if 'outlet' in var_in:
                cpi_menu.cpiexec.wait_for_threads()
            cpi_menu.print_attribute(var_in)
    else:
        # -- // LAUNCH MENU \\ --
        cpi_menu = ConsolePiMenu()
        while cpi_menu.go:
            cpi_menu.main_menu()
        print('hit')  # TODO remove debug line... This should never hit.
        cpi_menu.exit()



#!/etc/ConsolePi/venv/bin/python3

import sys
import yaml
from jinja2 import Environment, FileSystemLoader

sys.path.insert(0, '/etc/ConsolePi/src/pypkg')
from consolepi import log, utils # NoQA


# Look for config template by same name in same dir
if not len(sys.argv) > 1:
    print('Template Name Must be provided')
    sys.exit(1)
elif not utils.valid_file(sys.argv[1] + '.j2'):
    print(f'{sys.argv[1]}.j2 Not Found or empty')
    sys.exit(1)
else:
    cfg_templ = sys.argv[1] + '.j2'

# Look for variables file by same name in same dir
if utils.valid_file(sys.argv[1] + '.yaml'):
    var_file = sys.argv[1] + '.yaml'
# Look for single common variables file
elif utils.valid_file('./variables.yaml'):
    var_file = 'variables.yaml'
else:
    print('No Valid Variable File Found')
    sys.exit(1)

config_data = yaml.load(open(var_file), Loader=yaml.FullLoader)
if var_file == 'variables.yaml':
    config_data = config_data.get(sys.argv[1])


# Load Jinja2 template
env = Environment(loader=FileSystemLoader('./'), trim_blocks=True, lstrip_blocks=True)
template = env.get_template(cfg_templ)

# Render the template with data and print the output
print(template.render(config_data))



# generated by datamodel-codegen:
#   filename:  cloud.json
#   timestamp: 2022-10-02T03:48:01+00:00

# NOT Used just tinkering as I determine how best to re-factor / structure things

from __future__ import annotations
from pathlib import Path

from typing import Any, List, Mapping, Union

from pydantic import BaseModel


class SerialConfig(BaseModel):
    baud: int
    cmd: str
    dbits: int
    flow: str
    line: str
    log_ptr: Any
    logfile: Any
    parity: str
    port: int
    sbits: int


class Udev(BaseModel):
    devname: Path
    devpath: Path
    by_id: Path = None
    by_path: Path = None
    lame_devpath: str = None
    current_tags: str = None
    id_bus: str = None
    id_ifnum: str = None
    id_model: str = None
    id_model_from_database: str = None
    id_model_id: str = None
    id_path: str = None
    id_pci_class_from_database: str = None
    id_pci_subclass_from_database: str = None
    id_serial: str = None
    id_serial_short: str = None
    id_type: str = None
    id_usb_class_from_database: str = None
    id_usb_driver: str = None
    id_usb_interfaces: str = None
    id_usb_protocol_from_database: str = None
    id_vendor: str = None
    id_vendor_from_database: str = None
    id_vendor_id: str = None
    root_dev: bool = None
    subsystem: str = None
    time_since_init: str = None


class Adapter(BaseModel):
    config: SerialConfig
    outlets: List
    udev: Udev


class Interface(BaseModel):
    ip: str
    isgw: bool
    mac: str


class Remote(BaseModel):
    name: str
    adapters: Mapping[Path, Adapter]
    api_port: int
    cpuserial: str
    interfaces: Mapping[str, Union[Interface, str]]
    last_ip: str
    rem_ip: str = None
    source: str
    upd_time: int
    user: str
